/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.cache;

import java.io.File;
import java.util.Collection;
import java.util.Date;
import java.util.BitSet;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;

import casa.dodwan.docware.Descriptor;
import casa.dodwan.docware.DescriptorZipBufferizer;
import casa.dodwan.docware.Payload;
import casa.dodwan.docware.PayloadBufferizer;
import casa.dodwan.message.Message;
import casa.dodwan.util.FileBasedRepository;
import casa.dodwan.util.HybridRepository;
import casa.dodwan.util.Listenable;
import casa.dodwan.util.MemoryBasedRepository;
import casa.dodwan.util.Repository;
import casa.dodwan.util.Verbosity;
import casa.dodwan.util.DataFormatting;

public class Cache extends Listenable<CacheListener> {
    
    public class Fragments {
	// =======================================================
	// Maintains info about all fragments of a message that are
	// presently in the local cache.
	// =======================================================
	
	private BitSet has_;
	private int weight_ = 0;
	private int nbFragments_ = 0;

	// ----------------------------------------
	public Fragments(int nbFragments) {

	    nbFragments_ = nbFragments;
	    has_ = new BitSet(nbFragments);
	}

	// ----------------------------------------
	public int nbFragments() {
	    return nbFragments_;
	}

	// ----------------------------------------
	public int nbAvailableFragments() {
	    return has_.cardinality();
	}

	// ----------------------------------------
	public int nbMissingFragments() {
	    return nbFragments_ - has_.cardinality();
	}

	// ----------------------------------------
	public void setFragment(int fragId, int weight) {
	    if (! has_.get(fragId)) {
		has_.set(fragId);
		weight_ += weight;
	    }
	}

	// ----------------------------------------
	public boolean hasFragment(int fragId) {
	    return has_.get(fragId);
	}

	// ----------------------------------------
	public BitSet getFragments() {
	    return has_;
	}

	// ----------------------------------------
	public void clear() {
	    has_.clear();
	    weight_ = 0;
	    nbFragments_ = 0;
	}

	// --------------------------------------------------
	public int weight() {
	    return weight_;
	}
    }
    
    // ==================================================
    
    // Map of fragments
    protected Map<String, Fragments> fragments_ =
	new HashMap<String, Fragments>();
    
    // Repository for descriptors
    protected Repository<String, Descriptor> descriptors_;
    
    // Repository for payloads
    protected Repository<String, Payload> payloads_;
    
    private File path_ = null;
    
    private int weight_ = 0;
    
    protected PersistenceManager persistenceManager_;
    
    public Verbosity verbosity = new Verbosity();
    
    // ------------------------------------------------------------
    public Cache(File path) throws Exception {
	
	path_ = path;
	
	if (path_ == null) {
	    descriptors_ = new MemoryBasedRepository<String, Descriptor>();
	    payloads_ = new MemoryBasedRepository<String, Payload>();
	}
	else {
	    // 	    descriptors_ = new MemoryBasedRepository<String, Descriptor>();
	    descriptors_ = 
		new HybridRepository<Descriptor>(new File(path_, "descriptors"), 
						 new DescriptorZipBufferizer());
	    payloads_ = 
		new FileBasedRepository<Payload>(new File(path_, "payloads"), 
						 new PayloadBufferizer());
	}
	
	persistenceManager_ = new PersistenceManager(this);
    }
    
    // ------------------------------------------------------------
    public void load(String key, Descriptor desc) {
	
	try {
	    if (descriptors_ instanceof HybridRepository)
		((HybridRepository<Descriptor>)descriptors_).load(key, desc);
	    else
		descriptors_.put(key, desc);
	    scheduleRemoval(desc);
	    
	    int nbFrags = Integer.parseInt(desc.getAttribute("_nbfrags"));
	    Fragments frags = new Fragments(nbFrags);
	    fragments_.put(key, frags);

	    File ppath = new File(path_, "payloads");
	    for (int i=0; i < nbFrags; i++) {
		File file = new File(ppath, key + "-" + i);
		if (file.exists()) {
		    int length = (int)file.length();
		    frags.setFragment(i, length);
		    weight_ += length;
		}
	    }
	    
	} catch (Exception e) {
	    e.printStackTrace();
	}
	
	for (CacheListener listener : getListeners())
	    listener.addDescriptor(desc);
	
    }
    
    // ------------------------------------------------------------
    private void scheduleRemoval(Descriptor desc) {
	
	String key = desc.getDocumentId();
	
	if (key == null) {
	    System.out.println("Cache.scheduleRemoval(desc): abort (this document has no key)");
	    return;
	}
	
	try {
	    String ltime = desc.getAttribute("_ltime");
	    if (ltime != null) {
                long delay = Long.parseLong(ltime);
		persistenceManager_.schedule(key, delay);
		// casa.dodwan.util.Logger.log("cache", "lifetime=" + DataFormatting.to_hh_mm_ss(delay));
		if (verbosity.isEnabled())
		    System.out.println("  This document will be terminated in "
		     		       + DataFormatting.to_hh_mm_ss(delay)
		     		       + " (hh:mm:ss)");
		return;
	    }
	    Date deadline = desc.getDeadline();
	    if (deadline != null) {
		persistenceManager_.schedule(key, deadline);
		// casa.dodwan.util.Logger.log("cache", "deadline=" + deadline);
		if (verbosity.isEnabled())
		    System.out.println("  This document will be terminated on "	+ deadline);
		return;
	    }
	    
	}
	catch(Exception e) {
	    System.out.println("Cache.scheduleRemoval(desc): failed");
	}
	
	if (verbosity.isEnabled())
	    System.out.println("  This document requires no termination");
    }
    
    // ------------------------------------------------------------
    public synchronized void addDescriptor(Descriptor desc) {
	    
	    String key = desc.getDocumentId();
	    
	    if (verbosity.isEnabled()) {
		System.out.println("Cache.addDescriptor(" + key + ")");
	    }
	    
	    if (descriptors_.contains(key)) {
		if (verbosity.isEnabled())
		    System.out.println("  Abort: this document is already in the cache");
		return;
	    }
	    
	    try {
		casa.util.Logger.log("cache", "addDesc " + key);
		
		descriptors_.put(key, desc);
		int nbFrags = Integer.parseInt(desc.getAttribute("_nbfrags"));
		Fragments frags = new Fragments(nbFrags);

		fragments_.put(key, frags);
		
		scheduleRemoval(desc);
		
	    } catch (Exception e) {
		e.printStackTrace();
	    }
	    
	    for (CacheListener listener : getListeners())
		listener.addDescriptor(desc);
	}
    
    // ------------------------------------------------------------
    public synchronized void addPayload(String key, int fragId, 
					byte[] payload ) {
	    
	    if (verbosity.isEnabled()) {
		System.out.println("Cache.addPayload(" + key + ", " + fragId + ")");
	    }
	    
	    Descriptor desc = descriptors_.get(key);
	    if (desc == null) {
		if (verbosity.isEnabled())
		    System.out.println("  Abort: no descriptor in the cache for this payload");
		return;
	    }
	    
	    int nbFrags = Integer.parseInt(desc.getAttribute("_nbfrags"));
	    if (fragId >= nbFrags) {
		if (verbosity.isEnabled())
		    System.out.println("  Abort: invalid fragment id");
		return;
	    }
	    
	    try {
		Fragments frags = fragments_.get(key);
		if (frags.hasFragment(fragId)) {
		    if (verbosity.isEnabled())
			System.out.println("  Abort: this fragment is already in the cache");
		    return;
		}
		
		int ratio = 100 * (frags.nbAvailableFragments() + 1) / nbFrags;
		casa.util.Logger.log("cache", "addPayload " + key + "/" + fragId 
					    + " (" + ratio + "%)");
		
		payloads_.put(key + "-" + fragId, new Payload(payload));
		frags.setFragment(fragId, payload.length);
		weight_ += payload.length;
		
		for (CacheListener listener : getListeners())
		    listener.addPayload(key, fragId);
		
		if (frags.nbMissingFragments() == 0)
		    for (CacheListener listener : getListeners())
			listener.messageComplete(desc);	
	    } catch (Exception e) {
		e.printStackTrace();
	    }
	}
    
    // ------------------------------------------------------------
    public synchronized void cancel(Descriptor descriptor) throws Exception {
	    
	    cancel(descriptor.getDocumentId());
	}
    
    // ------------------------------------------------------------
    public synchronized void cancel(String key) throws Exception {
	
	if (verbosity.isEnabled())
	    System.out.println("Cache.cancel(" + key + ")");
	
	Descriptor desc = descriptors_.get(key);
	// Cannot cancel a message that does not exist
	if (desc == null) 
	    return;
	
	String status = desc.getAttribute("_cancelled");
	// Cannot cancel a message that has already been cancelled
	if ((status != null) && (status.equals("true")))
	    return;
	
	casa.util.Logger.log("cache", "cancel " + key);
	    
	desc.setAttribute("_cancelled", "true");
	descriptors_.put(key, desc);
    }
    
    // ------------------------------------------------------------
    public synchronized void purgePayload(String key) throws Exception {
	    
	if (verbosity.isEnabled())
	    System.out.println("Cache.purgePayload(" + key + ")");
	
	Descriptor desc = descriptors_.get(key);
	// Cannot delete the payload of a message that does not
	// exist
	if (desc == null) 
	    return;
	
	String status = desc.getAttribute("_cancelled");
	// Cannot delete the payload of a message that has not
	// been cancelled yet
	if ((status == null) || (! status.equals("true")))
	    return;
	
	// If nbFragments == 0, then this message has already been
	// purged (i.e., its fragments have been deleted from the
	// cache)
	int nbFrags = Integer.parseInt(desc.getAttribute("_nbfrags"));
	if (nbFrags == 0)
	    return;
	
	casa.util.Logger.log("cache", "purge " + key);
	
	// Let us replace the current descriptor by a very simple
	// (thus smaller) one
	Descriptor newDesc = new Descriptor();
	newDesc.setDocumentId(key);
	newDesc.setAttribute("_cancelled", "true");
        Date deadline = desc.getDeadline();
        if (deadline != null)
            newDesc.setDeadline(deadline);
	newDesc.setAttribute("_nbfrags", "0");
	descriptors_.put(key, newDesc);
	
	// Let us delete all fragments of the payload
	Fragments frags = fragments_.get(key);
	for (int i=0; i < frags.nbFragments();  i++) {
	    if (frags.hasFragment(i)) {
		payloads_.delete(key + "-" + i);
	    }
	}
	weight_ -= frags.weight();
	frags.clear();
    }
    
    // ------------------------------------------------------------
    public synchronized void remove(String key) throws Exception {
	    
	    if (verbosity.isEnabled())
		System.out.println("Cache.remove((" + key + "()");
	    
	    if (!descriptors_.contains(key)) {
		if (verbosity.isEnabled())
		    System.out
			.println("  Abort: this document is not in the cache");
		return;
	    }
	    
	    casa.util.Logger.log("cache", "remove " + key);
	    
	    // FG: temporary hack (notifying listeners that a message
	    // is going to be removed from the cache *before* it's
	    // actually removed)
	    for (CacheListener listener : getListeners())
		listener.remove(key);

	    persistenceManager_.cancel(key);
	    descriptors_.delete(key);
	    
	    Fragments frags = fragments_.get(key);
	    for (int i=0; i < frags.nbFragments();  i++) {
		if (frags.hasFragment(i)) {
		    payloads_.delete(key + "-" + i);
		}
	    }
	    weight_ -= frags.weight();
	    frags.clear();

	    // FG: former location of the notification to listeners
	    // (see temporary hask above)

	    // for (CacheListener listener: getListeners())
	    // listener.remove(key);
	}
    
    // ------------------------------------------------------------
    public synchronized void clear() throws Exception {

	casa.util.Logger.log("cache", "clear");

	Set<String> keys = new HashSet<String>();
	keys.addAll(getKeys());

	for (String key: keys)
	    remove(key);
    }

    // ------------------------------------------------------------
    public synchronized void update(Descriptor desc) throws Exception {
	
	String key = desc.getDocumentId();
	
	if (verbosity.isEnabled()) {
	    System.out.println("Cache.update('" + key + "')");
	}
	
	if (!descriptors_.contains(key)) {
	    if (verbosity.isEnabled())
		System.out
		    .println("  Abort: this document is not in the cache");
	    return;
	}
	
	casa.util.Logger.log("cache", "update " + key);
	
	descriptors_.put(key, desc);
	persistenceManager_.cancel(key);
	scheduleRemoval(desc);
	
	for (CacheListener listener : getListeners())
	    listener.update(desc);
    }
    
    // ------------------------------------------------------------
    public synchronized void update(Message message) throws Exception {
	    
	    update(message.getDescriptor());
	}
    
    // ------------------------------------------------------------
    public synchronized boolean contains(String key) {
	    
	    return descriptors_.contains(key);
	}
    
    // ------------------------------------------------------------
    public synchronized boolean contains(String key, int fragId) {
	    
	    Fragments frags = fragments_.get(key);
	    if (frags == null)
		return false;
	    else
		return frags.hasFragment(fragId);
	}
    
    // ------------------------------------------------------------
    public synchronized boolean isCancelled(String key) {
	    
	    Descriptor desc = getDescriptor(key);
	    if (desc == null)
		return false;

	    String status = desc.getAttribute("_cancelled");
	    if (status == null)
		return false;

	    return (status.equals("true"));
	}
    
    // ------------------------------------------------------------
    public synchronized Descriptor getDescriptor(String key) {
	    
	    return descriptors_.get(key);
	}
    
    // ------------------------------------------------------------
    public synchronized BitSet getFragments(String key) {
	    
	     Fragments frags = fragments_.get(key);
	     if (frags == null)
		 return null;
	     else 
		 return frags.getFragments();
	}
    
    // ------------------------------------------------------------
    public synchronized int getWeight() {
	    
	    return weight_;
	}
    
    // ------------------------------------------------------------
    public synchronized boolean isComplete(String key) {
	    
	    Fragments frags = fragments_.get(key);
	    if (frags == null)
		return false;
	    else
		return (frags.nbMissingFragments() == 0);
	}

    // ------------------------------------------------------------
    public synchronized int getWeight(String key) {
	    
	    Fragments frags = fragments_.get(key);
	    if (frags == null)
		return 0;

	    return frags.weight();

	    // int result = 0;
	    // if (payloads_ instanceof FileBasedRepository)
	    // 	result = (int)((FileBasedRepository<Payload>)payloads_).getSize(key);
	    // else {
	    // 	Payload payload = payloads_.get(key);
	    // 	if (payload != null)
	    // 	    result = payload.getBytes().length;
	    // }
	    
	    // return result;
	}
    
    // ------------------------------------------------------------
    public synchronized Payload getPayload(String key) {
	    
	    return payloads_.get(key);
	}
    
    // ----synchronized --------------------------------------------------------
    public Collection<Descriptor> getDescriptors() {
	// 	Collection<Descriptor> descriptors = new Vector<Descriptor>();
	// 	for (Message m : descriptors_.getElements())
	// 	    descriptors.add(m.getDescriptor());
	return descriptors_.getElements();
    }
    
    //     // ------------------------------------------------------------
    //     public Collection<Message> getElements() {
    
    // 	return descriptors_.getElements();
    //     }
    
    // ------------------------------------------------------------
    public synchronized Collection<String> getKeys() {
	    return descriptors_.getKeys();
	}
    
    // ------------------------------------------------------------
    public synchronized Collection<String> getKeys(Descriptor pattern) {
	    return pattern.getMatchingKeys(descriptors_.getElements());
	}
    
    // ------------------------------------------------------------
    public synchronized int getNbElements() {
	    return descriptors_.size();
	}
    
}
