/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.gossip;

import java.util.Date;
import java.util.Set;
import java.util.HashSet;
import java.util.BitSet;
import java.util.Collection;
import java.util.Iterator;
import java.util.Random;

import casa.dodwan.docware.Profile;
import casa.dodwan.docware.ProfileZipBufferizer;
import casa.dodwan.docware.Descriptor;
import casa.dodwan.docware.Payload;
import casa.dodwan.message.Message;
import casa.dodwan.cache.Cache;
import casa.dodwan.cache.CacheListener;

import casa.dodwan.util.Console;
import casa.dodwan.util.Service;
import casa.dodwan.util.Bufferizer;
import casa.dodwan.util.DataFormatting;
import casa.dodwan.util.Environment;
import casa.dodwan.util.Host;
import casa.dodwan.util.Catalog;
import casa.util.Logger;
import casa.dodwan.util.Sink;
import casa.dodwan.util.SystemEnvironment;
import casa.dodwan.util.Verbosity;
import casa.dodwan.util.Timer;
import casa.dodwan.util.RandomSeed;
import casa.dodwan.util.PresenceListener;
import casa.dodwan.config.DodwanEnvironment;

public class GossipingService extends Service
    implements Sink<Message>,CacheListener,PresenceListener {

    private Cache cache_;
    
    private Console console_ = null;
    
    public Environment environment;
    private Host localHost_ = null;
    private int offer_capacity_ = 1;

    private Message localProfile_;

    private Sink<Message> sink_;

    public Verbosity verbosity = new Verbosity();

    private Bufferizer<Profile> profileBufferizer_ = 
	new ProfileZipBufferizer();

    private Random random_ = new Random(RandomSeed.getSeed());

    private boolean degradedMode = false;
    
    // ------------------------------------------------------------
    /**
     * Create an instance of the GossipingService.
     * @param sink the sink for outgoing messages
     * @param cache the local cache where messages can be stored
     */
    public GossipingService(Sink<Message> sink,
			    Cache cache, 
			    Environment envt) {
	
	sink_ = sink;
	cache_ = cache;
	environment = envt;
	localHost_ = environment.getLocalHost();

	cache.addListener(this);

	offer_capacity_ = DodwanEnvironment.getInstance().offer_capacity;

	// Set a default empty context
	createProfileMessage();

	environment.addListener(this);

	String str = System.getProperty("degraded");
	try {
	    if (str != null)
		degradedMode = Boolean.parseBoolean(str);
	    if (degradedMode)
		Logger.log("gossip", "running in degraded mode");
	}
	catch (Exception e) {}
	
    }

    // ------------------------------------------------------------
    private void logPeerStatus(Host peer) {
	
	Logger.log("debug_gossip", "logPeerStatus: peer=" + peer.id
		   + ",couldUse={" + peer.couldUse
		   + "},canProvide={" + peer.canProvide
		   + "},duplicates={" + peer.duplicates + "}");

	// Logger.log("debug", "peer=" + peer.id);
	// Logger.log("debug", "  couldUse={" + peer.couldUse + "}");
	// Logger.log("debug", "  canProvide={" + peer.canProvide + "}");
	// Logger.log("debug", "  duplicates={" + peer.duplicates + "}");
    }
    
    // ------------------------------------------------------------
    private void sendRequest(Host peer,
			     String ackId, Catalog denied) {

	try {
	    // Logger.log("debug", "sendRequest(peer=" + peer.id
	    // 	       + ",ackId=" + ackId
	    // 	       + ", denied=" + denied + ")");
	    // logPeerStatus(peer);
	     
	// Let us identify messages available on that peer and
	// that are marked as 'cancelled' in the local cache: these
	// messages shall be signalled in the next request
	Set<String> cids = new HashSet<String>();

	// synchronized(peer) {
	    
	    for (String key: peer.canProvide.keySet()) 
		if (cache_.isCancelled(key)) 
		    cids.add(key);
	    // FG: direct
	    Set<String> dup = new HashSet<String>(peer.duplicates.keySet());
	    // for (String key: peer.duplicates.keySet())
	    for (String key: dup)
		if (cache_.isCancelled(key)) 
		    cids.add(key);
	    // FG: ---------
	    peer.canProvide.remove(cids);
	    peer.duplicates.remove(cids);
	    // } 

	// Let us identify duplicate messages, that is, messages that
	// are offered by that peer and that are already in the
	// local cache. These messages shall be denied in the next
	// request
	Catalog duplicates = new Catalog();
	// Logger.log("debug", "Searching for duplicates");
	for (String key: peer.canProvide.keySet()) {
	    if (cache_.contains(key)) {
		BitSet localFrags = cache_.getFragments(key);
		BitSet canProvide = peer.canProvide.getFragments(key);
		// FG: devrait plus arriver...
		// if (canProvide.isEmpty())
		//     // If this peer can only provide a descriptor
		//     // (and no payload fragment), let us signal that
		//     // we already have this descriptor in the local
		//     // cache
		//     denied.put(key);
		// else {
		    // If this peer can provide payload
		    // fragments, let us only signal those fragments
		    // we already have in the local cache
		BitSet dupFrags = (BitSet)canProvide.clone();
		    dupFrags.and(localFrags);
		    if (! dupFrags.isEmpty())
			duplicates.put(key, dupFrags);
		// }
	    }
	}
	// Logger.log("debug", "duplicates=" + duplicates);
	// logPeerStatus(peer);

	// Update peer.duplicates by adding duplicates that have just
	// been identified: there is no need to offer them to the
	// peer
	// Logger.log("debug", "Updating peer.duplicates");
	peer.duplicates.add(duplicates);
	// Logger.log("debug", "  peer.duplicates=" + duplicates);

	// Update peer.canProvide by removing duplicates that have
	// just been identified: there is no need to request them from
	// the peer
	// Logger.log("debug", "Updating peer.canProvide");
	peer.canProvide.remove(duplicates);
	// Logger.log("debug", "  peer.canProvide=" + peer.canProvide);

	
	// // Remove messages we want to deny from the peer's
	// // 'available' set
	// if (denied != null) {
	//     Logger.log("debug", "Removing denied from peer.canProvide");
	//     peer.canProvide.remove((Catalog)denied.clone());
	//     Logger.log("debug", "peer.canProvide=" + peer.canProvide);
	//     Logger.log("debug", "denied=" + denied);
	// }

	// Combine the messages we have identified as duplicates
	// (which must be denied) with the 'denied' set (provided as a
	// parameter of that method)
	// Logger.log("debug", "Combining duplicates and denied");
	if (denied == null)
	    denied = new Catalog();
	denied.add((Catalog)duplicates.clone());
	// Logger.log("debug", "  denied=" + denied);

	if ((ackId != null) && (denied != null)) {
	    // Let us make sure the message we are ack'ing (if any) is
	    // not mentioned in denied
	    //
	    // 'ackId' should be of the form 'xxxx/yy'
	    // Logger.log("debug", "Removing ackId from denied");
	    int idx = ackId.lastIndexOf("/");
	    String msgKey = ackId.substring(0, idx);
	    String str2 = ackId.substring(idx+1);
	    int fragId = Integer.parseInt(str2);
	    BitSet frags = denied.getFragments(msgKey);
	    if (frags != null) {
		frags.clear(fragId);
		if (frags.isEmpty())
		    denied.remove(msgKey);
	    }
	    // Logger.log("debug", "  denied=" + denied);
	}

	// There is no need to transport an empty 'denied' set in the
	// request if we can avoid it
	if ((denied != null) && denied.isEmpty()) {
	    denied = null;
	    // Logger.log("debug", "  denied=" + denied);
	}
	
	// If the peer can still provide messages, then let us
	// select the message we are going to request first
	// Logger.log("debug", "Selecting the next message to request");
	String mid = null;
	if (! peer.canProvide.isEmpty()) {
	    // Let us select the message randomly, in order to better
	    // distribute requests between several peers
	    String[] selectable = new String[0];
	    selectable = peer.canProvide.keySet().toArray(selectable);
	    mid = selectable[random_.nextInt(selectable.length)];
	}
	    
	if (mid == null) {
	    // If no message has been selected, then let us signal the
	    // peer that there is nothing left we can request
	    // (based on the current offer)
	    mid = "_none_";
	    peer.expected = null;
	}
	else {
	    // Let us select a payload fragment we are missing for
	    // message 'mid'
	    BitSet offeredFrags = peer.canProvide.getFragments(mid);
	    int fragId;
	    if (offeredFrags.get(0))
		// If fragment 0 is missing, let us request that fragment
		fragId = 0;
	    else {
		// Otherwise let us select randomly one of the other
		// missing fragments
		int idx = random_.nextInt(offeredFrags.cardinality());
		fragId = offeredFrags.nextSetBit(idx);
	    }
	    mid += "/" + fragId;
	    peer.expected = mid;
	}
	// Logger.log("debug", "  mid=" + mid);
	

	// Logger.log("debug", "Sending request");

	// Let us create and send the request
	Message request = createRequest(peer.id, mid,
					ackId, denied, 
					cids, null, null);
	sendMessage(request);

	}
	catch(Exception e) {
	    e.printStackTrace();
	}
    }

    // ------------------------------------------------------------
    private boolean descMatchesProfile(Descriptor desc,
				       Set<Descriptor> profile) {

	for (Descriptor pattern: profile) 
	    if (desc.matches(pattern))
		return true;
	return false;
    }


    // ------------------------------------------------------------
    private void sendMessage(Message msg) {

	try {
	    // Send offer
	    sink_.write(msg);
	}
	catch (Exception e) {
	    Logger.log("gossip", "warning: tx failed");
	    e.printStackTrace();
	}	
    }

    // ------------------------------------------------------------
    private boolean obsolete(Message msg) {

	// A peer can send us an obsolete message if its clock is
	// late compared to ours. In that case the message's remaining
	// lifetime (according to our clock) is negative

	Descriptor desc = msg.getDescriptor();
	Date deadline = desc.getDeadline();
	if (deadline == null)
	    return false;

	long now = System.currentTimeMillis();
	long lifetime = deadline.getTime() - now;

	return (lifetime <= 0);
    }

    // ------------------------------------------------------------
    private void processStandardMessage(String key,
					String sdr,
					Message msg) {

	// Let us check if this message was expected from that
	// peer. If the peer is unknown, or if a different
	// message is expected from that peer, then the message
	// we've just received is considered as an 'alien' message
	Host peer = environment.getPeer(sdr);
	boolean alien = ((peer == null) || (! key.equals(peer.expected)));

	// 'key' should be of the form 'xxxx/yy'
	int idx = key.lastIndexOf("/");
	String msgKey = key.substring(0, idx);
	String str2 = key.substring(idx+1);
	int fragId = Integer.parseInt(str2);

	boolean obsolete = false;
	boolean duplicate = false;

	if (! alien) {
	    // Let us check if this message is obsolete
	    if (fragId == 0)
		// Message obsolescence can only be checked if the
		// message contains a descriptor
		obsolete = obsolete(msg);
		
	    // Let us check if this message is a duplicate
	    duplicate = cache_.contains(msgKey, fragId);
	}
	
	Logger.log("provide", "< sdr=" + sdr
		   + ",mid=" + key
		   + (obsolete?" (obsolete)":"")
		   + (duplicate?" (dup)":"")
		   + (alien?" (alien)":""));
		   
	// If the message is an alien message, then let us discard it
	if (alien)
	    return;

	// If the message is obsolete, then let us make sure we won't
	// request it again (or any other fragment of this message for
	// that matter) from the same peer
	if (obsolete) {
	    BitSet frags = peer.canProvide.getFragments(msgKey);
	    Catalog denied = new Catalog();
	    denied.put(msgKey, (BitSet)frags.clone());
	    Logger.log("debug", "denying obsolete message: {"
		       + denied + "}");
	    // Logger.log("debug", "Updating peer.canProvide");
	    peer.canProvide.remove(msgKey);
	    // Logger.log("debug", "  peer.canProvide:" + peer.canProvide);
	    sendRequest(peer, null, denied);
	    return;
	}

	// If the message is not a duplicate and is not obsolete, let
	// us store it in the local cache
	if (! duplicate && ! obsolete) {

	    // If we have received the first fragment of message
	    // 'msgKey', then let us store its descriptor in the cache
	    if (fragId == 0) {
		Descriptor desc = msg.getDescriptor();
		desc.setDocumentId(msgKey);
		cache_.addDescriptor(desc);
	    }
	    
	    // Let us store the payload in the cache
	    cache_.addPayload(msgKey, fragId, msg.getPayload());
	}
		   
	// Let us send another request to that peer, while
	// confirming we've received this message it sent us
	sendRequest(peer, key, null);
     }
    
    // ------------------------------------------------------------
    private Message createRequest(String rcv,
				  String mid, 
				  String mack,
				  Catalog denied,
				  Set<String> cancelIds,
				  String pid,
				  String pack) {

	// Logger.log("debug", "createRequest(denied=" + denied + ")");
	
	Message msg = new Message();
	Descriptor desc = new Descriptor();
	msg.setDescriptor(desc);
	msg.setPayload(new byte[0]);

	desc.setUniqueDocumentId();
	
	desc.setAttribute("_rcv", rcv);

	desc.setAttribute("_gtype", "request");

	if (mid != null) 
	    desc.setAttribute("_mid", mid);

	if (mack != null)
	    desc.setAttribute("_mack", mack);

	String str_dids = null;
	if ((denied != null) && !denied.isEmpty()) {
	    str_dids = DataFormatting.toString(denied, ",");
	    desc.setAttribute("_dids", str_dids);
	}

	String str_cids = null;
	if ((cancelIds != null) && !cancelIds.isEmpty()) {
	    str_cids = DataFormatting.toString(cancelIds, ",");
	    desc.setAttribute("_cids", str_cids);
	}

	if (pid != null)
	    desc.setAttribute("_pid", pid);

	if (pack != null)
	    desc.setAttribute("_pack", pack);
	

	Logger.log("request", 		   "> rcv=" + rcv
		   + (mid!=null?",mid="+mid:"")
		   + (mack!=null?",mack="+mack:"")
		   + (str_dids!=null?",dids={"+str_dids+"}":"")
		   + (str_cids!=null?",cids={"+str_cids+"}":"")
		   + (pid!=null?",pid="+pid:"")
		   + (pack!=null?",pack="+pack:""));

	return msg;
    }

    // ------------------------------------------------------------
    private void processProfile(String sdr,
				Descriptor desc, 
				byte[] payload) {

	String str_pid = desc.getAttribute("_pid");

	if (str_pid == null) {
	    Logger.log("error", "No pid in profile");
	    return;
	}

	int pid = Integer.parseInt(str_pid);
	
	Host peer = environment.getPeer(sdr);

	// If this message was received from an unknown peer, let
	// us discard it
	if (peer == null) {
	    Logger.log("error", "Unknown peer");
	    return;
	}

	boolean dup = (pid == peer.pid);

	Logger.log("profile", "< sdr=" + peer.id 
		   + ",pid=" + pid
		   + (dup?" (dup)":" (update)"));

	if (! dup) {
	    // The peer's profile needs to be updated
	    Profile prof = null;
	    try {
		prof = profileBufferizer_.fromBuffer(payload);
	    }
	    catch (Exception e) {
		Logger.log("error", "Could not extract profile from payload");
		return;
	    }
	    Set<Descriptor> descriptors = prof.descriptors();

	    peer.pid = pid;
	    environment.setPeerProfile(peer.id, descriptors);
	}

	// Let us send an acknowledgement to the peer
	Message request = createRequest(peer.id, "_none_", null,
					null, null, null, str_pid);
	sendMessage(request);
	
	// If no transaction is engaged with that peer, let us
	// try to send an offer
	if (peer.busy.compareAndSet(false, true)) {
	    Logger.log("gossip", "id=" + peer.id
		       + ", lock acquired");
	    updateAndSendOfferForPeer(peer);
	}
    }

    // ------------------------------------------------------------
    private void updateAndSendOfferForPeer(Host peer) {

	String pid = null;
	if (localHost_.pid != peer.lastAckedPid) {
	    // If we suspect this peer does not have our current
	    // profile, let us offer to send this profile
	    pid = String.valueOf(localHost_.pid);
	    peer.couldUse.clear();
	}
	else
	    // Otherwise let us try to select messages we can offer to
	    // that peer
	    updateOfferForPeer(peer);

	// Logger.log("offer", "id=" + peer.id 
	// 	   + ",pid=" + pid
	//  	   + ",#mids=" + peer.couldUse.size());

	 // If there is something we can propose to that peer
	 // (either message fragments, or our current profile), let us
	 // do so
	if (! (peer.couldUse.isEmpty() && (pid == null)))
	    sendOffer(peer, pid);
	else {
	    // Otherwise let us record that no transaction is
	    // currently engaged with that peer
	    Logger.log("gossip", "id=" + peer.id
		       + ", lock released");
	    peer.busy.set(false);
	}   
    }

    // ------------------------------------------------------------
    private BitSet checkAvailFragments(String key, Host peer) {

	// Consider offering message 'key' to that peer.  This is
	// only possible for fragments of the message that are not
	// already in peer.canProvide (1) or in peer.duplicates (2);
	
	BitSet availFrags = (BitSet)cache_.getFragments(key).clone();
	BitSet offeredFrags = peer.canProvide.getFragments(key);
	if (offeredFrags != null)
	    availFrags.andNot(offeredFrags); // (1)
	BitSet dupFrags = peer.duplicates.getFragments(key);
	if (dupFrags != null)
	    availFrags.andNot(dupFrags); // (2)

	return availFrags;
    }

    // ------------------------------------------------------------
    private Set<String> getDestinations(Descriptor desc) {

	// Returns the destinations of this message (if any) 
	String dst = desc.getAttribute("_dst");
	return DataFormatting.splitToSet(dst, ",");
    }

    // ------------------------------------------------------------
    private Set<String> getReceived(Descriptor desc) {

	// Returns the destinations this message has already reached
	// (if any)
	String rcvd = desc.getAttribute("_rcvd");
	return DataFormatting.splitToSet(rcvd, ",");
    }

    // ------------------------------------------------------------
    private void cancelMessage(String key) {

	Descriptor desc = cache_.getDescriptor(key);
	
	try {
	    Logger.log("gossip", "cancel mid=" + key);
	    cache_.cancel(key);

	    // If the message had no set destinations, or if the
	    // local host was *not* one of these destinations,
	    // then the message's payload can be purged from the
	    // cache
	    Set<String> rcvd = getReceived(desc);
	    if ((rcvd == null) || (! rcvd.contains(localHost_.id))) {
		// Logger.log("gossip", "purge mid=" + key + " [auto]");
		cache_.purgePayload(key);
	    }
	}
	catch (Exception e) {
	    Logger.log("gossip", "cancel " + key + " failed");
	}
    }

    // ------------------------------------------------------------
    private void updateOfferForPeer(Host peer) {

	// Logger.log("offer", "(update) id=" + peer.id);

	// Let us make sure we hold the lock for that peer
	if (! peer.busy.get())
	    Logger.log("error", "id=" + peer.id
		       + ", unlocked");

	// Let us make sure no other thread modifies the local cache
	// while we are parsing it
	synchronized(cache_) {

	    // Let us first clear the previous offer
	    peer.couldUse.clear();
	    
	    // Let us build an offer for that peer based on
	    // messages available in the local cache, while making
	    // sure that each message considered in this offer is not
	    // already available on the peer, has not been
	    // cancelled yet, matches the peer's interest profile,
	    // and keeping in mind that the number of message
	    // identifiers in an offer is limited (pfiuuu!)
	    
	    int cnt = 0;
	    
	    // Let us make sure that all descriptors and payload
	    // fragments considered as duplicates (between this
	    // peer and the local host) are still present in the
	    // local cache
	    Set<String> lost = new HashSet<String>();
	    for (String key: peer.duplicates.keySet()) {
		if (cache_.contains(key)) {
		    BitSet dupFrags = peer.duplicates.getFragments(key);
		    BitSet localFrags = cache_.getFragments(key);
		    dupFrags.and(localFrags);
		}
		else
		    lost.add(key);
	    }
	    // synchronized(peer) {
		peer.duplicates.remove(lost);
		// }

	    // For each document in the local cache, let us verify...
	    // 
	    // (1) that it has not been cancelled yet;
	    // (2) that it matches the peer's interest profile
	    // (3) that it cannot be directly sent to another peer;
	    // (4) that at least some fragments of the document can be
	    // offered to the peer.
	    
	    Iterator<String> iter = cache_.getKeys().iterator();
	    while (iter.hasNext() && (cnt < offer_capacity_)) {
		String key = iter.next();
		
		if (! cache_.isCancelled(key)) {  // (1)

		    Descriptor desc = cache_.getDescriptor(key);
		    if (descMatchesProfile(desc, peer.profile)) {  // (2)

			Set<String> dst = getDestinations(desc);
			// Logger.log("debug_gossip", "updateOfferForPeer: mid=" + key
			// 	   + ",dst={" + dst + "}");

			// FG: direct
			boolean mayOffer = mayOffer(key, peer.id, dst);
			if (mayOffer) {  // (3)

			    BitSet availFrags = 
				checkAvailFragments(key, peer);
			    
			    if (! availFrags.isEmpty()) {  // (4)
				peer.couldUse.put(key, availFrags);
				cnt++;
			        // cnt += availFrags.cardinality();
			    }
			}
		    }
		}
	    }
	}

	// logPeerStatus(peer);

	// Logger.log("offer", "id=" + peer.id 
	// 	   + ",#mids=" + peer.couldUse.size());
    }

    // ------------------------------------------------------------
    private void sendOffer(Host peer, String pid) {
	
	Message msg = new Message();
	Descriptor desc = new Descriptor();
	msg.setDescriptor(desc);
	msg.setPayload(new byte[0]);

	desc.setUniqueDocumentId();
	desc.setAttribute("_rcv", peer.id);
	desc.setAttribute("_gtype", "offer");
	if (pid != null) {
	    desc.setAttribute("_spid", pid);
	    desc.setAttribute("_rpid", String.valueOf(peer.pid));
	}
	
	String mids = null;
	Catalog couldUse = peer.couldUse;
	if (! couldUse.isEmpty()) {
	    mids = DataFormatting.toString(couldUse, ",");
	    desc.setAttribute("_mids", mids);
	}

	// Logging this offer
	Logger.log("offer", 
		   "> rcv=" + peer.id
		   + (pid != null?",spid=" + pid + ",rpid=" + peer.pid:"")
		   + (mids != null?",mids={" + mids + "}":""));
	
	sendMessage(msg);

	// Register this transaction event
	peer.helloCnt = 0;
    }

    // ------------------------------------------------------------
    private void sendMissingNotification(Host peer, 
					 String mid,
					 String cid) {

	Message msg = new Message();
	Descriptor desc = new Descriptor();
	msg.setDescriptor(desc);
	msg.setPayload(new byte[0]);

	desc.setUniqueDocumentId();
	desc.setAttribute("_rcv", peer.id);
	desc.setAttribute("_gtype", "missing");

	if (mid != null) {
	    desc.setAttribute("_mid", mid);
	    peer.couldUse.remove(mid);
	}
	
	if (cid != null)
	    desc.setAttribute("_cid", cid);

	// Logging this notification
	Logger.log("missing", 
		   "> rcv=" + peer.id
		   + (mid != null?",mid=" + mid:"")
		   + (cid != null?",cid=" + cid:""));
	sendMessage(msg);
    }

    // ------------------------------------------------------------
    private void processOffer(String sdr, Descriptor desc) {
	
	String str_spid  = desc.getAttribute("_spid");
	String str_rpid  = desc.getAttribute("_rpid");
	String mids = desc.getAttribute("_mids");

	// Logger.log("debug", "< offer rcvd from " + sdr
	// 	   + " with sdr=" + desc.getAttribute("_sdr")
	// 	   + ",rcv=" + desc.getAttribute("_rcv"));
	
	Logger.log("offer",
		   "< sdr=" + sdr 
		   + (str_spid != null?",spid=" + str_spid:"")
		   + (str_rpid != null?",rpid=" + str_rpid:"")
		   + (mids != null?",mids={" + mids + "}":""));

	Host peer = environment.getPeer(sdr);
	if (peer == null) {
	    // FG
	    // {
	    //     synchronized(environment) {
	    //      	peer = environment.addPeer(sdr);
	    //     }
	    // }
	    Logger.log("offer",
		       "discarding: sdr=" + sdr + " is unknown");
	    return;
	}


	Catalog offer;
	if (mids != null)
	    offer = DataFormatting.splitToCatalog(mids, ",");
	else
	    offer = new Catalog();

	// FG
	// Record that this peer owns those messages
	// peer.duplicates.addAll(offer);

	// FG: pb ci-dessous. Il faudrait identifier tous les messages
	// offerts par le voisin et qui sont deja presents dans le
	// cache local. En fait c'est fait dans sendRequest, mais ce
	// n'est peut-etre pas le meilleur
	// endroit. checkPeerRequest examine peer.duplicates,
	// alors qu'il faudrait aussi examiner peer.canProvide (ou
	// plus exactement l'union des deux).
	
	// FG
	// For each message offered by that peer, check that
	// the descriptor of this message (contained in fragment 0) is
	// either part of the offer, or is already available
	// locally. Any message that does not satisfy these
	// requirements should be denied.
	Catalog denied = new Catalog();
	for (String key: offer.keySet()) {
	    BitSet frags = offer.getFragments(key);
	    if (! cache_.contains(key, 0) && ! frags.get(0))
		denied.put(key, (BitSet)frags.clone());
	}
	// Logger.log("debug", "denied=" + denied + ", offer=" + offer);
	if (denied.isEmpty())
	    denied = null;
	else {
	    // Logger.log("debug", "Removing denied from offer");
	    offer.remove(denied);
	    // Logger.log("debug", "denied=" + denied + ", offer=" + offer);
	}
	
	// Record that offered messages can be requested from that peer
	peer.canProvide = offer;
	// Logger.log("debug", "peer.canProvide=" + peer.canProvide);

	// FG: direct
	//
	// For each message the peer has offered, let us check if
	// it was a destination of that message. If so, then maybe we
	// can register that that peer is not a destination for
	// that message anymore
	for (String key: offer.keySet())
	    checkPeerReached(peer, key);
	
	if (str_rpid != null) {
	    // Check that the peer is up-to-date with our current
	    // profile, and if that's not the case let us try to offer
	    // our profile
	     int rpid = Integer.parseInt(str_rpid);
	     peer.lastAckedPid = rpid;
	     if (rpid != localHost_.pid)
		 if (peer.busy.compareAndSet(false, true)) {
		     Logger.log("gossip", "id=" + peer.id
				+ ", lock acquired");
		     updateAndSendOfferForPeer(peer);
		 }
	}

	if (str_spid != null) {
	    // If the peer has offered to send its profile, let us
	    // consider this option
	    int spid = Integer.parseInt(str_spid);
	    Message request = null;
	    if (spid != peer.pid) {
		// If we do not have the peer's current profile
		// yet, let us request this profile
		Logger.log("profile", 
			   "> request rcv=" + peer.id
			   + ",pid=" + spid);

		request = createRequest(peer.id, null, null, 
					null, null, str_spid, null);
	    }
	    else {
		// Otherwise let us acknowledge that we already have
		// this profile
		request = createRequest(peer.id, "_none_", null, 
					null, null, null, str_spid);
	    }
	    sendMessage(request);
	    return;
	}

	// Let us send a request to that peer, based on what the
	// peer has offered
	sendRequest(peer, null, denied);
    }

    // ------------------------------------------------------------
    private void processMissingNotification(String sdr,
					    Descriptor desc) {

	// Let us process a notification message that informs us that
	// a message we have apparently requested is not available
	// anymore

	// Let us first extract the message ids from the notification
	String mid = desc.getAttribute("_mid");
	String cid = desc.getAttribute("_cid");

	Host peer = environment.getPeer(sdr);
	boolean alien = (peer == null);
	Logger.log("missing", "< sdr=" + sdr
		   + (mid != null?",mid=" + mid:"")
		   + (cid != null?",cid=" + cid:"")
		   + (alien?" (alien)":""));

	// If the peer is unknown, then the
	// notification we've just received is considered as an
	// 'alien' message and it should be ignored
	if (alien)
	    return;

	// If a message must be cancelled, let us do it and send
	// another request to that peer
	if (cid != null) {
	    try {
		cancelMessage(cid);

		// synchronized(peer) {
		    peer.canProvide.remove(cid);
		    peer.duplicates.remove(cid);
		    // }
	    }
	    catch (Exception e) {
		Logger.log("gossip", "warning: failed to cancel " + cid);
	    }
	    sendRequest(peer, null, null);
	    return;
	}

	if (mid == null) {
	    Logger.log("error", "No mid in notification");
	    return;
	}
	
	// Let us check the specified message was expected from that
	// peer. If not, that's strange but we'll do the job
	// anyway
	boolean expected = (! mid.equals(peer.expected));

	if (! expected) 
	    Logger.log("gossip", "warning: cancel notification received for unexpected msg " 
		       + mid);

	// Let us record that this message is not available on that
	// peer
	Catalog missing = DataFormatting.splitToCatalog(mid, ",");

	// synchronized(peer) {
	    peer.canProvide.remove(missing);
	    peer.duplicates.remove(missing);
	    // }

	// Let us send another request to that peer, but only if
	// this message was expected from that peer
	sendRequest(peer, null, null);
    }

    // ------------------------------------------------------------
    private void processRequest(String sdr, 
				Descriptor desc) {

	String mid = desc.getAttribute("_mid");
	String str_cids = desc.getAttribute("_cids");
	String mack  = desc.getAttribute("_mack");
	String str_dids = desc.getAttribute("_dids");
	String str_pid = desc.getAttribute("_pid");
	String pack = desc.getAttribute("_pack");

	// Logging this request
	Logger.log("request", 
		   "< sdr=" + sdr
		   + (mid!=null?",mid="+mid:"")
		   + (mack!=null?",mack="+mack+"":"")
		   + (str_dids!=null?",dids={"+str_dids+"}":"")
		   + (str_cids!=null?",cids={"+str_cids+"}":"")
		   + (str_pid!=null?",pid="+str_pid:"")
		   + (pack!=null?",pack="+pack+"":""));
	
	Host peer = environment.getPeer(sdr);
	if (peer == null) {
	    Logger.log("gossip", "warning: discarding request received from unknown peer " 
		       + sdr);
	    return;
	    // synchronized(environment) {
	    // 	peer = environment.addPeer(sdr);
	    // }
	}

	// Register this transaction event
	peer.helloCnt = 0;
	
	// If the peer has acked the profile we have sent
	// previously, let us accound for this event
	if (pack != null) {
	    int ackedPid = -1;
	    ackedPid = Integer.parseInt(pack);
	    peer.lastAckedPid = ackedPid;
	}

	// If the peer has acked some message fragment we have
	// sent previously, let us account for this event
	if (mack != null) {
	    int idx = mack.lastIndexOf("/");
	    String msgKey;
	    msgKey = mack.substring(0, idx);
	    String str2 = mack.substring(idx+1);
	    int fragId = Integer.parseInt(str2);

	    // synchronized(peer) {
		peer.duplicates.set(msgKey, fragId);
		    peer.couldUse.clear(msgKey, fragId);
		    // }	
	    
	    // FG: direct
	    //
	    // Let us check if the peer was a destination of the
	    // message it has just acknowledged. If so, then maybe we
	    // can register that that peer is not a destination
	    // for that message anymore
	    checkPeerReached(peer, msgKey);
	}
    

	// If the peer has denied some of the messages we have
	// offered, let us record that it already owns these messages
	// and that we must not offer them again
	if (str_dids != null) {
	    Catalog denied = DataFormatting.splitToCatalog(str_dids, ",");

	    // synchronized(peer) {
		peer.duplicates.add(denied);
		peer.couldUse.remove(denied);
		// }

	    // FG: direct
	    //
	    // For each message the peer has denied, let us check
	    // if it was a destination of that message. If so, then
	    // maybe we can register that that peer is not a
	    // destination for that message anymore
	    for (String key: denied.keySet())
	     	checkPeerReached(peer, key);
	}
	
	// If the peer has signalled that some of the messages we
	// have offered have been cancelled, let us cancel these
	// messages locally
	if (str_cids != null) {

	    // synchronized(peer) {
		for (String key: str_cids.split(","))
		    try {
			if (! cache_.isCancelled(key))
			    cancelMessage(key);
			peer.canProvide.remove(key);
			peer.duplicates.remove(key);
		    }
		    catch (Exception e) {
			Logger.log("gossip", "warning: failed to cancel "
				   + key);
		    }
		// }
	}

	// logPeerStatus(peer);

	// If the peer requests a copy of our local interest
	// profile, let us send this profile
	if (str_pid != null) {
	    Descriptor desc2 = localProfile_.getDescriptor();
	    int pid = Integer.parseInt(str_pid);
	    if (pid != localHost_.pid) {
		Logger.log("profile", "warning: pid=" + pid
			   + " different from current one (" + localHost_.pid
			   + ")");
		pid = localHost_.pid;
	    }
	    Logger.log("profile", "> rcv=" + peer.id 
		       + ",pid=" + pid);
	    
	    desc2.setAttribute("_rcv", peer.id);
	    sendMessage(localProfile_);
	}

	// If the request does not contains the '_mid' attribute, then
	// there is nothing left to do
	if (mid == null)
	    return;
	
	// Any request for a message should be received while a
	// transaction is engaged with that peer
	if (peer.busy.compareAndSet(false, true)) {
	    Logger.log("warning", "< id=" + peer.id
		       + ", unlocked");
	    Logger.log("gossip", "< id=" + peer.id
		       + ", lock acquired");
	}

	if (mid.equals("_none_")) {
	    // The peer signalled that there is nothing it
	    // can request based on our last offer. Let us try to
	    // update this offer
	    
	    // Let us try to select new messages we can offer to that
	    // peer
	    updateAndSendOfferForPeer(peer);
	}
	
	else {
	    // The peer has requested a message fragment. Let us
	    // try to get this fragment from the cache and send it to
	    // the peer

	    // 'mid' should be of the form 'xxxx/yy'
	    int idx = mid.lastIndexOf("/");
	    String msgKey = mid.substring(0, idx);
	    String str2 = mid.substring(idx+1);
	    int fragId = Integer.parseInt(str2);

	    // Let us send a reply to the peer
	    sendPayload(msgKey, fragId, mid, peer);
	}
    }
 
    // ------------------------------------------------------------
    private void processTime(String sdr, 
			     Descriptor desc) {

	long now  = System.currentTimeMillis();
	String st = desc.getAttribute("st");
	long sent = Long.parseLong(st);
	long offset = now - sent;
	
	// Logging this timing info
	Logger.log("time", 
		   "< sdr=" + sdr
		   + ",sent=" + sent
		   + ",now=" + now
		   + ",offset=" + offset);
    }
    
    // ------------------------------------------------------------
    private void sendPayload(String key, int fragId, String mid,
			     Host peer) {

	if (cache_.isCancelled(key)) {
	    sendMissingNotification(peer, null, key); 
	    return;
	}

	if (! cache_.contains(key, fragId)) {
	    sendMissingNotification(peer, mid, null); 
	    return;
	}

	Payload payload = cache_.getPayload(key + "-" + fragId);

	if (payload != null) {
	    Descriptor desc = null;
	    // If we are preparing to send fragment 0, then let us
	    // include a copy of the global message's descriptor with
	    // that payload. Otherwise let us build a short descriptor
	    // from scratch.
	    if (fragId == 0) 
		desc = new Descriptor(cache_.getDescriptor(key));
	    else
		desc = new Descriptor();
	    desc.setDocumentId(key + "/" + fragId);
	    desc.setAttribute("_rcv", peer.id);
	    Message msg = new Message(desc);
	    msg.setPayload(payload.getBytes());
	    Logger.log("provide", "> " + "rcv=" + peer.id
		       + ",mid=" + mid);
	    sendMessage(msg);
	}
	else
	    // Let us inform the peer that the requested
	    // descriptor is not available locally
	    sendMissingNotification(peer, mid, null);  
    }

    // ------------------------------------------------------------
    public void processMessage(Message message) throws Exception {

	Descriptor desc = message.getDescriptor();
	
	String key = desc.getDocumentId();
	String sdr = desc.getAttribute("_sdr");
	
	if (sdr == null) {
	    Logger.log("error", "No sdr in message");
	    return;
	}
	
	if (verbosity.isEnabled())
	    System.out.println("GossipingService.write('"
			       + key + "')");
	
	// If this message is not a gossiping control message, then
	// let us process it as a standard message
	String gossipType = desc.getAttribute("_gtype");
	if (gossipType == null) {
	    processStandardMessage(key, sdr, message);
	    return;
	}
	
	// Let's try not to process our own service messages
	if (sdr.equals(SystemEnvironment.host)) {
	    if (verbosity.isEnabled())
		System.out.println(" Abort: ignoring our own announces");
	    return;
	}

	// Any message that contains indication about the sender's
	// time should be processed
	if (gossipType.equals("time")) {
	    processTime(sdr, desc);
	    return;
	}

	// Oterwise let us filter out messages that have not been
	// addressed to us explicitly
	String rcv = message.getDescriptor().getAttribute("_rcv");
	if ((rcv == null)
	    || (! rcv.equals(SystemEnvironment.host)))
	    return;

	// Logger.log("debug", "< msg rcvd with sdr=" + sdr
	// 	   + ",rcv=" + rcv);

	if (gossipType.equals("profile")) {
	    processProfile(sdr, desc, message.getPayload());
	    return;
	}

	if (gossipType.equals("offer")) {
	    processOffer(sdr, desc);
	    return;
	}

	if (gossipType.equals("missing")) {
	    processMissingNotification(sdr, desc);
	    return;
	}
	if (gossipType.equals("request")) {
	    processRequest(sdr, desc);
	    return;
	}
    }

    // ------------------------------------------------------------
    public void write(Message message) throws Exception {

	if (isSuspended())
	    return;
	  
	// synchronized(this) {

	    processMessage(message);

	    // }
    }

    // ------------------------------------------------------------
    private void initiateTransactions() {

	Set<Host> currentPeers = new HashSet<Host>();
	synchronized(environment) {
	    currentPeers.addAll(environment.getPeers());
	}
	
	for (Host peer: currentPeers) 
	    if (peer.busy.compareAndSet(false, true)) {
		Logger.log("gossip", "id=" + peer.id
			   + ", lock acquired");
		updateAndSendOfferForPeer(peer);
	    }
    }

    // ============================================================
    // Implementation of direct transmissions for destination-based
    // messages
    // ============================================================
    
    private void checkPeerReached(Host peer, String key) {

	// FG: direct
	
	// Let us check if message 'key' has a destination set, if
	// peer 'peer' is part of this set, and if all fragments
	// of the message have already reached that peer. If that
	// is the case, then let us register that that peer is not
	// a destination for the message anymore.
	
	// Logger.log("gossip", "checkPeerReached id=" + peer.id
	//  	   + ",mid=" + key);
	
	Descriptor desc = cache_.getDescriptor(key);
	if (desc == null)
	    return;

	// Let us make sure message 'key' has not been cancelled. If
	// so, then let us return right now
	String status = desc.getAttribute("_cancelled");
	if ((status != null) && (status.equals("true")))
	    return;
	
	// Let us first check if message 'key' has a destination set,
	// and if peer 'peer' is part of this set. If not, then
	// there is nothing we can do
	Set<String> dst = getDestinations(desc);
	if ((dst == null) || (! dst.contains(peer.id)))
	    return;

	// Logger.log("debug_gossip", "checkPeerReached: dst={" + dst + "} contains " + peer.id);

	// Let us now check if the peer already owns all the
	// fragments of this message
	int nbFrags = Integer.parseInt(desc.getAttribute("_nbfrags"));
	if (nbFrags <= 0)
	    return;
	BitSet dup_frags = peer.duplicates.getFragments(key);
	BitSet avail_frags = peer.canProvide.getFragments(key);
	// Logger.log("gossip", "checkPeerReached id=" + peer.id
	// 	   + ",mid=" + key
	// 	   + ": " + fragments.cardinality() + "/ " + nbFrags);
	int nbDup = 0;
	if (dup_frags != null)
	    nbDup = dup_frags.cardinality();
	if (avail_frags != null)
	    nbDup += avail_frags.cardinality();
	int nbMissing = nbFrags - nbDup;
	if (nbMissing != 0) {
	    // Logger.log("gossip", "id=" + peer.id
	    // 	       + " still missing "
	    // 	       + nbMissing + " fragments for mid=" + key);
	    return;
	}
	else {
	    // Logger.log("gossip", "id=" + peer.id
	    // 	       + " owns all fragments for mid=" + key);
	}
	
	// The peer is not a destination for the message
	// anymore. Let us update its descriptor's destination
	// attribute accordingly
	removeDestination(desc, peer.id);
    }
    
    // ------------------------------------------------------------
    private boolean mayOffer(String key, String rcv, Set<String> dest) {

	// FG: direct
	
	// Determine if msg 'key' with destination 'dest' should be
	// offered to peer 'rcv'

	// Logger.log("direct", ">? mid=" + key
	// 	   + ",id=" + rcv + ",dst={"
	// 	   + DataFormatting.toString(dest, ",")
	// 	   + "}");

	// If 'dest' is not specified, then epidemic dissemination is
	// required, so the msg should be offered to any peer
	if (dest == null) {
	    // Logger.log("direct", "dst=null => epidemic dissemination");
	    return true;
	}

	// Let us remove the local host's id from the set of
	// destinations. If there are no remaining destinations after
	// that, then there is no need to offer the message to anybody
	dest.remove(localHost_.id);
	if (dest.isEmpty()) {
	    // Logger.log("direct", "dst={" + localHost_.id +
	    // 	       "} => no need to disseminate further");
	    return false;
	}

	// If the message is addressed to the peer considered
	// (possibly among other destinations), then of course it
	// should be offered to that peer
	if (dest.contains(rcv)) {
	    // Logger.log("direct", "host " + rcv + " is a destination");
	    return true;
	}

	// If all remaining destinations are currently peers of
	// the local host, then direct transfer to these peers is
	// possible and there is no need to offer the message to any
	// node (the case where the peer considered is itself a
	// destination has been dealt with in the above statement)
	if (environment.getIds().containsAll(dest)) {
	    // Logger.log("direct", "dest={" + DataFormatting.toString(dest, ",")
	    // 	       + "} are all peers => direct transfer is possible");
	    return false;
	}

	// In any other case, let us offer the message to that
	// peer
	// Logger.log("direct", "default behavior: epidemic dissemination");
	return true;
    }

    // ------------------------------------------------------------
    private void checkDestination(String key) {

	// FG: direct

	Descriptor desc = cache_.getDescriptor(key);
	if (desc == null)
	    return;
	
	// Let us check if this message has already been cancelled
	// (possibly by another Listener<Cache>). If so then there is
	// nothing to do...
	String status = desc.getAttribute("_cancelled");
	if ((status != null) && (status.equals("true")))
	    return;

	// Let us check if this message has a set destination. If not
	// then there is nothing we can do..;
	// Set<String> dst = getDestination(desc);
	// Logger.log("debug_gossip", "checkDestination: mid=" + mid
	// + ",dst={" + dst + "}");
	// if (dst == null)
	//     return;

	// If the local host was a destination of this message, then
	// let us remove the local host's id from the descriptor's
	// destination attribute
	removeDestination(desc, localHost_.id);
    }

    // ------------------------------------------------------------
    private void removeDestination(Descriptor desc, String hostId) {

	// FG: direct
	
	String key = desc.getDocumentId();
	
	Set<String> dst = getDestinations(desc);
	if ((dst == null) || (! dst.contains(hostId)))
	    return;
	
	Set<String> rcvd = getReceived(desc);
	if (rcvd == null)
	    rcvd = new HashSet<String>();
	
	dst.remove(hostId);
	rcvd.add(hostId);
	String str_dst = DataFormatting.toString(dst, ",");
	String str_rcvd = DataFormatting.toString(rcvd, ",");
	desc.setAttribute("_dst", str_dst);
	desc.setAttribute("_rcvd", str_rcvd);

	Logger.log("gossip", "update mid=" + key
		   + ":dst={" + str_dst + "},rcvd={" + str_rcvd + "}");
	
	// If the message has no remaining destination, then it can be
	// cancelled automatically
	if (dst.isEmpty())
	    cancelMessage(key);
    }

    // ============================================================
    // Implementation of the CacheListener interface
    // ============================================================

    public void addDescriptor(Descriptor desc) {

	// Nothing to do
    }

    // ------------------------------------------------------------
    public void addPayload(String key, int fragId) {

	if (isSuspended())
	    return;

	// A message fragment has just been put in the cache.

	// If the corresponding message is now complete, let us check
	// if the local host was a destination of this message. If so
	// then maybe the message can be be cancelled automatically
	if (cache_.isComplete(key))
	    checkDestination(key);
	
	// This new fragment may be of interest to some peers. Let
	// us try to initiate transactions with peers with which
	// transactions are not already engaged.
	if (degradedMode) return;
	
	//	synchronized(this) {
	    initiateTransactions();
	    // }
    }

    // ------------------------------------------------------------
    public void messageComplete(Descriptor desc) {
    }

    // ------------------------------------------------------------
    public void cancel(String key) {
    }
    
    // ------------------------------------------------------------
    public void remove(String key) {
    }
    
    // ------------------------------------------------------------
    public void update(Descriptor desc) {
    }


    // ============================================================
    // Management of the local profile
    // ============================================================

    // ------------------------------------------------------------
    private void createProfileMessage() {
	
	casa.util.Logger.log("profile",
				    "(create)");

	// Let us set a random value for the local profile's id
	localHost_.pid = random_.nextInt(100000);
	
	// Let us create the profile message and its descriptor
	localProfile_ = new Message();
	Descriptor desc = new Descriptor();
	localProfile_.setDescriptor(desc);
	desc.setAttribute("_gtype", "profile");
	desc.setAttribute("_pid", String.valueOf(localHost_.pid));
	desc.setUniqueDocumentId();

	casa.util.Logger.log("profile",
				    "create (local) pid=" + localHost_.pid);
    }

    // ------------------------------------------------------------
    private void updateProfileMessage() {
	
	Descriptor desc = localProfile_.getDescriptor();
	
	localHost_.pid = (localHost_.pid + 1) % 1000;
	String str_pid = String.valueOf(localHost_.pid);
	desc.setAttribute("_pid", str_pid);

	desc.setUniqueDocumentId();

	Profile prof = new Profile();
	for (Descriptor pattern : localHost_.profile) 
	    prof.add(pattern);

	try {
	    byte[] payload = profileBufferizer_.toBuffer(prof);
	    localProfile_.setPayload(payload);
	}
	catch (Exception e) {
	    e.printStackTrace();
	    return;
	}

	casa.util.Logger.log("profile",
				    "update (local) pid=" + localHost_.pid);
    }

    // ------------------------------------------------------------
    public synchronized void addToLocalProfile(Descriptor pattern) {
	    
	localHost_.profile.add(pattern);
	updateProfileMessage();

	// The local profile has just been been modified. This
	// information should be transmitted to peers as soon as
	// possible. Let us try to initiate transactions with
	// peers with which transactions are not already engaged.
	initiateTransactions();
    }
    
    // ------------------------------------------------------------
    public synchronized void removeFromLocalProfile(Descriptor pattern) {
	    
	localHost_.profile.remove(pattern);
	updateProfileMessage();
	
	// The local profile has just been been modified. This
	// information should be transmitted to peers as soon as
	// possible. Let us try to initiate transactions with
	// peers with which transactions are not already engaged.
	initiateTransactions();
    }


    // ============================================================
    // Implementation of the PresenceListener interface
    // ============================================================

    private void checkPendingTransaction(Host peer) {

	// Let us check if a very old transaction is engaged with that
	// peer. If so, then let us try to re-engage this
	// transaction
	if (peer.busy.compareAndSet(true, true)) {
	    // There is indeed an ongoing transaction...
	    if (peer.helloCnt >= 2) {
		// ... but the last request from this peer was
		// long, long ago (which means the current transaction
		// has failed somewhow)
		Logger.log("gossip", "warning: a transaction with "
			   + peer.id + " was engaged long ago");
		updateAndSendOfferForPeer(peer);
	    }
	}
    }

    // ------------------------------------------------------------
    public void addPeer(String id) {

	if (isSuspended())
	    return;

	// synchronized(this) {
	    
	    // Let us log this announce
	    Logger.log("gossip", "add " + id);
	    
	    Host peer = environment.getPeer(id);
	    // Let us make sure we propose our current profile to that
	    // peer in the next offer
	    peer.lastAckedPid = -1;
	    peer.helloCnt = 0;
	    
	    // Let us check there is no pending transaction with that
	    // peer
	    checkPendingTransaction(peer);
	    
	    // Since this peer is apparently a new one, then let us try
	    // to engage a transaction with that peer
	    if (peer.busy.compareAndSet(false, true)) {
		Logger.log("gossip", "id=" + id
			   + ", lock acquired");

		if (degradedMode)
		    // Forget everything we know about this peer's catalog
		    peer.duplicates.clear();
		
		updateAndSendOfferForPeer(peer);
	    }

	    // }
    }
    
    // ------------------------------------------------------------
    public void removePeer(String id) {

	if (isSuspended())
	    return;

	// synchronized(this) {
	
	    Logger.log("gossip", "remove " + id);
	    
	    Host peer = environment.getPeer(id);
	    if (peer == null)
		return;
	    
	    // If a transaction was engaged with that peer, let us cancel
	    // it
	    if (peer.busy.compareAndSet(true, false)) 
		Logger.log("gossip", "id=" + id
			   + ", lock released");

	    // }
    }

    // ------------------------------------------------------------
    public void updatePeer(String id) {

	if (isSuspended())
	    return;

	// synchronized(this) {
	    
	    // Let us log this announce
	    // Logger.log("gossip", "update " + id);
	    
	    Host peer = environment.getPeer(id);
	    if (peer == null) {
		Logger.log("gossip", "warning: unkown peer=" + id);
	    return;
	    }
	    
	    peer.helloCnt++;
	    
	    // Let us check there is no pending transaction with that
	    // peer
	    checkPendingTransaction(peer);

	    // }
    }

    // ------------------------------------------------------------
    public void clearPeers() {

	if (isSuspended())
	    return;

	// Let us log this announce
	// Logger.log("gossip", "clear");
    }

    // ============================================================
    // Implementation of the Service interface
    // ============================================================

    public void suspend() {
	
	if (verbosity.isEnabled())
	    System.out.println("GossipingService.suspend()");
	
	if (isSuspended_) {
	    if (verbosity.isEnabled())
		System.out.println("  Abort: already suspended...");
	    return;
	}

	super.suspend();
    }
    
    // ------------------------------------------------------------
    public void resume() {
	
	if (verbosity.isEnabled())
	    System.out.println("GossipingService.resume()");
	
	super.resume();
    }

    // ------------------------------------------------------------
    public Console console() {
	if (console_ == null)
	    console_ = new GossipingServiceConsole(this);
	return console_;
    }
    
    // ------------------------------------------------------------
    public void sendTime(String target) {

	Descriptor desc = new Descriptor();
	desc.setUniqueDocumentId();
	desc.setAttribute("_gtype", "time");
	if (target != null)
	    desc.setAttribute("_rcv", target);
	Message msg = new Message();

	msg.setDescriptor(desc);
	msg.setPayload(new byte[0]);
	long now = System.currentTimeMillis();
	desc.setAttribute("st", String.valueOf(now));
	sendMessage(msg);

	Logger.log("time", 
		   "> rcv=" + target
		   + ",st=" + now);
    }
}
