/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/


package casa.dodwan.hub;

import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.util.concurrent.CyclicBarrier;

import casa.util.Logger;
import casa.dodwan.util.DataFormatting;
import casa.util.DataBufferizer;
import casa.dodwan.net.Stream_IO;
import casa.dodwan.net.StreamSender;
import casa.dodwan.message.Message;
import casa.dodwan.docware.Descriptor;
import casa.dodwan.message.MessageBufferizer;

public class SocketRelay {

    // public static int pipeCounter = 0;
    // public static int activePipes = 0;
    // private static long start = System.currentTimeMillis();
    
    // ==================================================
    private class StreamPipe extends Thread {

	private Socket inSocket_;
	private Socket outSocket_;
	private SocketRelay relay_;
	public String peerId = null;

	boolean running = false;
	
	public StreamPipe(Socket in,
			  Socket out,
			  SocketRelay relay)
	throws Exception {

	    inSocket_ = in;
	    outSocket_ = out;
	    relay_ = relay;
	}
	
	// ------------------------------------------------------------
	private void parseHandshake(InputStream in,
				      OutputStream out) 
	    throws Exception {
	    
	    // The first two bytes should contain a cookie
	    byte[] b_cookie = Stream_IO.receiveBytes(in, 2);
	    // short cookie = DataBufferizer.readShort(b_cookie, 0);
	    
	    // Now comes the peer's id
	    byte[] b_id = Stream_IO.receiveBundle(in);
	    peerId = DataBufferizer.readString(b_id, 0);

	    // Waiting until the other pipe has also parsed its cookie
	    // and peer's id
	    relay_.barrier.await();

	    // Let us now check that both pipes can be connected
	    boolean connected = relay_.checkConnectivity();
	    if (! connected) {
		Logger.log("relay", "Discarding connection");
		peerId = null;
		return;
	    }
	    
	    Stream_IO.sendBytes(out, b_cookie);
	    Stream_IO.sendBundle(out, b_id);
	}
	
	// --------------------------------------------------
	private void runBytewise(InputStream in,
				 OutputStream out)
	    throws Exception {

	    byte[] buffer = new byte[1024];
	    int len = in.read(buffer);
	    while (len != -1) {
		relay_.checkConnectivity();
		
		// Logger.log("relay",
		// 	       "[" + len + "]> "
		// 	       + this);
		out.write(buffer, 0, len);
		len = in.read(buffer);
	    }
	}
	
	// --------------------------------------------------
	private void runMessagewise(InputStream in,
				    OutputStream out)
	    throws Exception {

	    MessageBufferizer bufferizer = 
		new MessageBufferizer();

	    synchronized(this) {
		while (true) {
		    byte[] buffer = Stream_IO.receiveBundle(in);
		    Logger.log("relay", "< [" + buffer.length + "] via " + this);
		    relay_.checkConnectivity();
		    
		    Message msg = bufferizer.fromBuffer(buffer);
		    Descriptor desc = msg.getDescriptor();
		    String simtag = desc.getAttribute("_simtag");
		    if (simtag != null) {
			String sdr   = desc.getAttribute("_sdr");
			String rcv   = desc.getAttribute("_rcv");
			relay_.processSimTag(sdr, rcv, simtag);
		    }
		    Logger.log("relay", "  > " + msg.getDescriptor());
		    StreamSender sender = new StreamSender(out, buffer);
		    sender.run();
		}
	    }
	}
	
	// --------------------------------------------------
	public void run() {

	    InputStream in = null;
	    OutputStream out = null;
	    try {
		in = inSocket_.getInputStream();
		out = outSocket_.getOutputStream();
	    }
	    catch(Exception e) {
		Logger.log("relay", "error: failed to start " + this);
		return;
	    }
	    
	    running = true;
		
	    try {
		parseHandshake(in, out);
 		if (peerId == null) {
		    this.close();
		    return;
		}
		
		// pipeCounter++;
		// activePipes++;
		// this.setName("pipe-" + pipeCounter);
		// System.err.println("Start " + this.getName()
		// 		   + " (active=" + activePipes + ")");
		// long now = System.currentTimeMillis() - start;
		// System.err.println(now + " " + activePipes);
		// Logger.log("pipe", "peer=" + peerId);

		relay_.notifyConnectionStatus(true);
		
		// runBytewise(in, out);
		runMessagewise(in, out);
	    }
	    catch (Exception e) {
	    }

	    this.close();
	}

	// --------------------------------------------------
	public void close() {

	    running = false;
	    
	    if (relay_ != null) {
		// activePipes--;
		// long now = System.currentTimeMillis() - start;
		// System.err.println(now + " " + activePipes);
		relay_.close();
		inSocket_ = null;
		outSocket_ = null;
		relay_ = null;
		// System.err.println("Terminate " + this.getName()
		// 		   + " (active=" + activePipes + ")");
	    }
	}
	
	// --------------------------------------------------
	public String toString() {

	    return "Pipe("
		+ DataFormatting.toString((InetSocketAddress)inSocket_.getRemoteSocketAddress())
		+ " => "
		+ DataFormatting.toString((InetSocketAddress)outSocket_.getRemoteSocketAddress())
		+ ")";
	}
	
    }
    // ==================================================

    private Socket sock1_;
    private StreamPipe pipe1_;
    private Socket sock2_;
    private StreamPipe pipe2_;
    private boolean isClosed_ = false;
    private Environment env_;
    
    protected CyclicBarrier barrier; // Barrier used to synchronize the two pipes
    
    public SocketRelay(Socket sock1, Socket sock2,
		       Environment env)
	throws Exception {

	sock1_ = sock1;
	sock2_ = sock2;
	env_ = env;
	barrier = new CyclicBarrier(2);
	
	Logger.log("relay", "creating " + this);
	pipe1_ = new StreamPipe(sock1_, sock2_, this);
	pipe2_ = new StreamPipe(sock2_, sock1_, this);

	pipe1_.start();
	pipe2_.start();

    }

    // ------------------------------------------------------------
    public synchronized void close() {

	if (isClosed_)
	    return;

	isClosed_ = true;
	
	Logger.log("relay", "closing " + this);
	
	try { sock1_.close(); } catch(Exception e) {}
	try { sock2_.close(); } catch(Exception e) {}

	// System.err.println("DISCONNECTED: " + pipe1_.peerId
	// 		   + " " + pipe2_.peerId);
	// env_.setConnectionStatus(pipe1_.peerId, pipe2_.peerId,
	// false);
	// FG: 
	// notifyConnectionStatus(false);

    	sock1_ = null;
	pipe1_ = null;
	sock2_ = null;
	pipe2_ = null;
}

    // ------------------------------------------------------------
    public synchronized void notifyConnectionStatus(boolean online) {

	if ((pipe1_.peerId != null) && (pipe2_.peerId != null)) {
	    // System.err.println("CONNECTED: " + pipe1_.peerId
	    // 		       + " " + pipe2_.peerId);
	    env_.setConnectionStatus(pipe1_.peerId, pipe2_.peerId,
				     online);
	}
    }
    
    // ------------------------------------------------------------
    public synchronized boolean checkConnectivity() {

	boolean connected = env_.arePeers(pipe1_.peerId, pipe2_.peerId);
	if (! connected)
	    close();

	return connected;
    }

    // ------------------------------------------------------------
    public synchronized void processSimTag(String sdr, String rcv,
					   String simtag) {

	Logger.log("relay", "tagging " + sdr + " and " + rcv);
	env_.setNodeStatus(sdr, simtag);
	env_.setNodeStatus(rcv, simtag);
    }

    // ------------------------------------------------------------
    public String toString() {

	return "SocketRelay("
	    + (pipe1_==null?"":pipe1_.peerId + "/")
	    + DataFormatting.toString((InetSocketAddress)sock1_.getRemoteSocketAddress())
	    + " <=> "
	    + (pipe2_==null?"":pipe2_.peerId + "/")
	    + DataFormatting.toString((InetSocketAddress)sock2_.getRemoteSocketAddress())
	    + ")";
    }
    
}
