/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.message;

import casa.dodwan.docware.Descriptor;
import casa.dodwan.util.Keyable;

/**
 * DoDWAN message. A Dodwan message is composed of descriptor and a
 * payload.  The content-based dissemination of DoDWAN messages is
 * based only on the descriptor's content.  The payload is an array of
 * bytes that is not interpreted by DoDWAN.
 */
public class Message implements Keyable<String> {
    
    private Descriptor descriptor_ = null;
    
    private byte[] payload_ = null;
    
    // ------------------------------------------------------------
    /**
     * Creates an empty message. This message contains no descriptor and no payload.
     */
    public Message() {
    }
    

    // ------------------------------------------------------------
    /**
     * Creates a message with an initial descriptor. This message contains no payload.
     * @param descriptor the initial descriptor.
     */
    public Message(Descriptor descriptor) {
	descriptor_ = descriptor;
    }
    
    // ------------------------------------------------------------
    /**
     * Creates a message from an initial descriptor and payload.
     * @param descriptor the initial descriptor.
     * @param payload the initial payload.
     */
    public Message(Descriptor descriptor, byte[] payload) {
	descriptor_ = descriptor;
	payload_ = payload;
    }
    
    // ------------------------------------------------------------
    /**
     * Creates a message as the deep clone of another specified message. The descriptor 
     * (resp. payload) of this message is a (deep) clone of that of the specified 
     * descriptor (resp. payload).
     * @param other the other message that must be cloned.
     */
     public Message(Message other) {

	descriptor_ = new Descriptor(other.descriptor_);
	
	if (other.payload_ == null) {
	    this.payload_ = null;
	    return;
	}
	
	this.payload_ = new byte[other.payload_.length];
	System.arraycopy(other.payload_, 0, this.payload_, 0,
			 other.payload_.length);
    }
    
    // ------------------------------------------------------------
    /**
     * Returns a (deep) clone of this message.
     * @return a (deep) clone of this message.
     */
     public Object clone() {
	
	Message clone = new Message();

	clone.descriptor_ = new Descriptor(descriptor_);

	if (this.payload_ == null) 
	    clone.payload_ = null;
	else {
	    clone.payload_ = new byte[this.payload_.length];
	    System.arraycopy(this.payload_, 0, clone.payload_, 0,
			     this.payload_.length);
	}

	return clone;
    }
    
    // ------------------------------------------------------------
    /**
     * Returns the key that can serve as a unique identifier for this
     * message. This key is the one of the descriptor embedded in this message.
     * @return the key that can serve as a unique identifier for this
     *         message.
     */
     public String getKey() {
	return descriptor_.getDocumentId();
    }
    
    // ------------------------------------------------------------
    /**
     * Returns the payload of this message.
     * @return the payload of this message.
     */
     public byte[] getPayload() {
	return payload_;
    }
    
    // ------------------------------------------------------------
    /**
     * Returns the descriptor of this message.
     * @return the descriptor of this message.
     */
    public Descriptor getDescriptor() {
	return descriptor_;
    }
    
    // ------------------------------------------------------------
    /**
     * Sets the payload (or change the existing one) of this message.
     * @param payload the new value for the payload 
     */
    public void setPayload(byte[] payload) {
	payload_ = payload;
    }
    
    // ------------------------------------------------------------
    /**
     * Sets the descriptor (or change the existing one) of this message.
     * @param descriptor the new value for the descriptor 
     */
    public void setDescriptor(Descriptor descriptor) {
	descriptor_ = descriptor;
    }


    /**
     * Returns a string representation of the message.
     * Output is : <br>
     * Message : <br>
     * descriptor attribute name  1 : descriptor attribute value 1 <br>
     * descriptor attribute name  2 : descriptor attribute value 2 <br>
     * descriptor attribute name  3 : descriptor attribute value 3 <br>
     * ... <br>
     * Payload size
     * @return String representation of the message
     */
    @Override
    public String toString() {
        String ret = "Message :\n";

        for (String name: descriptor_.getNames()) {
            ret += name + " : " + descriptor_.getAttribute(name) + "\n";
        }
        ret += ("Payload + " + (payload_ == null ? "null" : payload_.length));

        return ret;
    }
}
