/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.net;

import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.Collection;

import casa.dodwan.docware.Descriptor;
import casa.dodwan.message.Message;
import casa.dodwan.message.MessageBufferizer;
import casa.dodwan.util.PresenceListener;
import casa.util.Logger;
import casa.util.Processor;
import casa.dodwan.util.DataFormatting;
import casa.dodwan.config.DodwanEnvironment;

public class ChannelManager
    implements ChannelListener,PeerListener {
    
    private PresenceListener listener_;

    private Map<String,TCP_Server> servers_ = 
	new HashMap<String,TCP_Server>();

    private Map<String,Channel> channels_ = 
	new HashMap<String,Channel>();

    private Map<String,Channel> peers_ = 
	new HashMap<String,Channel>();

    private Processor<Message> processor_ = null;

    private int channelCounter = 0;
    
    // ------------------------------------------------------------
    public ChannelManager(PresenceListener listener,
			  Processor<Message> processor) {
	
	listener_ = listener;
	processor_ = processor;
    }


    // ============================================================
    // Server management
    // ============================================================

    public boolean hasServer(String serverId) {

	return servers_.containsKey(serverId);
    }

    // ------------------------------------------------------------
    public Map<String,TCP_Server> getServers() {
	
	return servers_;
    }

    // ------------------------------------------------------------
    public synchronized void addServer(String serverId,
				       TCP_Server server) {

	if (servers_.containsKey(serverId)) {
	    Logger.log("chan_mgr", "warning: add server " + serverId
		       + " failed (server already registered)");
	    return;
	}

	Logger.log("chan_mgr", "add " + serverId
		   + "=" + server);
	servers_.put(serverId, server);
    }

    // ------------------------------------------------------------
    public synchronized String getServerId(TCP_Server server) {
	
	Iterator<String> iter = servers_.keySet().iterator();
	while (iter.hasNext()) {
	    String id = iter.next();
	    if (server.equals(servers_.get(id))) {
		return id;
	    }
	}
	return null;
    }

   // ------------------------------------------------------------
    public synchronized void removeServer(TCP_Server server) {

	String id = getServerId(server);

	if (id != null)
	    removeServer(id);
    }

    // ------------------------------------------------------------
    public synchronized void removeServer(String serverId) {

	TCP_Server server = servers_.get(serverId);

	if (server == null) 
	    return;

	Logger.log("chan_mgr", "remove server " + serverId
		   + "(" + server + ")");
	
	server.close();

	servers_.remove(serverId); 
    }

    // ============================================================
    // Channel management
    // ============================================================

    public synchronized void addChannel(Channel channel) {

	if (channels_.containsValue(channel)) {
	    Logger.log("chan_mgr", "warning: add channel failed (channel already registered)");
	    return;
	}

	String channelId = "ch%" + channelCounter++;

	addChannel(channelId, channel);
    }

    // ------------------------------------------------------------
    public synchronized void addChannel(String channelId,
					Channel channel) {

	if (channels_.containsKey(channelId)) {
	    Logger.log("chan_mgr", "warning: add " + channelId
		       + " failed (channel already registered)");
	    return;
	}

	Logger.log("chan_mgr", "add " + channelId
		   + "=" + channel);
	channels_.put(channelId, channel);
    }

    // ------------------------------------------------------------
    public synchronized String getChannelId(Channel channel) {
	
	Iterator<String> iter = channels_.keySet().iterator();
	while (iter.hasNext()) {
	    String id = iter.next();
	    if (channel.equals(channels_.get(id))) {
		return id;
	    }
	}
	return null;
    }
    
    // ------------------------------------------------------------
    public synchronized void removeChannel(String channelId) {

	Channel channel = channels_.get(channelId);

	if (channel == null) 
	    return;

	Logger.log("chan_mgr", "remove " + channelId
		   + "(" + channel + ")");
	
	if (channel.isOpen())
	    channel.close();

	channels_.remove(channelId); 
    }

   // ------------------------------------------------------------
    public synchronized void removeChannel(Channel channel) {

	String id = getChannelId(channel);

	if (id != null)
	    removeChannel(id);
    }

    // ------------------------------------------------------------
    private void closeChannels() {

	Set<String> channelIds = new HashSet<String>();
	channelIds.addAll(channels_.keySet());

	for (String id: channelIds) {
	    Channel channel = channels_.get(id);
	    if ((channel != null) && channel.isOpen())
		channel.close();
	}
    }

    // ------------------------------------------------------------
    public synchronized void clear() {

	Logger.log("chan_mgr", "clear");

	// Let us try to cleanly close all channels
	closeChannels();

	// Let us do it again (!), in case new channels appeared while
	// we were closing older ones
	closeChannels();

	// If all channels have not disappeared nicely, let us forget
	// the remaining ones (this should not happen, but who
	// knows...)
	if (! channels_.isEmpty()) {
	    String chan_str = DataFormatting.toString(channels_.keySet(), ",");
	    Logger.log("chan_mgr", "warning: purging remaining channels ("
		       + chan_str + ")");
	    channels_.clear();
	}

	// Likewise all peers should have been removed nicely by now,
	// but just is case...
	if (! peers_.isEmpty()) {
	    String peer_str = DataFormatting.toString(peers_.keySet(), ",");
	    Logger.log("chan_mgr", "warning: purging remaining peers ("
		       + peer_str + ")");
	    
	    if (listener_ != null)
		for (String peerId: peers_.keySet()) 
		    listener_.removePeer(peerId);
	    peers_.clear();
	}
    }

    // ------------------------------------------------------------
    public boolean hasChannel(String channelId) {

	return channels_.containsKey(channelId);
    }

    // ------------------------------------------------------------
    public Channel getChannel(String channelId) {
	
	return channels_.get(channelId);
    }
    
    // ------------------------------------------------------------
    public Channel getChannelForPeer(String peerId) {
	
	return peers_.get(peerId);
    }
    
    // ------------------------------------------------------------
    public long getLastRxForPeer(String peerId) {

	Channel ch =  peers_.get(peerId);
	if (ch != null)
	    return ch.lastRx();
	else
	    return -1;
    }
    
    // ------------------------------------------------------------
    public Map<String,Channel> getChannels() {
	
	return channels_;
    }
    
     // ------------------------------------------------------------
    public boolean hasPeer(String peerId) {

	return peers_.containsKey(peerId);
    }

    // ------------------------------------------------------------
    public Set<String> getPeers() {
	
	return peers_.keySet();
    }
    
    // ============================================================
    // Implementation of the ChannelListener interface
    // ============================================================

    public synchronized void channelCreated(Channel channel) {

	addChannel(channel);
    }
    
    // ------------------------------------------------------------
    public synchronized void channelClosed(Channel channel) {

	removeChannel(channel);
    }
    
    // ============================================================
    // Implementation of the PeerListener interface
    // ============================================================

    public synchronized void peerDiscovered(String peerId, Channel channel) {

	Logger.log("chan_mgr", "add peer "
		   + peerId
		   + " (via " + channel + ")");

	boolean newPeer = (! peers_.containsKey(peerId));
	peers_.put(peerId, channel);

	if (listener_ != null)
	    if (newPeer)
		listener_.addPeer(peerId);
	    else
		listener_.updatePeer(peerId);
    }

    // ------------------------------------------------------------
    public synchronized void peerLost(String peerId) {

	Channel channel = peers_.get(peerId);
	
	if (channel == null)
	    return;
	
	Logger.log("chan_mgr", "remove peer "
		   + peerId);

	peers_.remove(peerId);

	if (listener_ != null)
	    listener_.removePeer(peerId);
    }

    // ------------------------------------------------------------
    public synchronized void peerConfirmed(String peerId) {

	// Logger.log("chan_mgr", "update peer "
	// 	   + peerId);

	if (! peers_.containsKey(peerId)) {
	    Logger.log("chan_mgr", "warning: unknown peer "
		       + peerId);
	    return;
	}

	if (listener_ != null)
	    listener_.updatePeer(peerId);
    }

    // ------------------------------------------------------------
    public void process(Message msg) {

	if (processor_ == null) {
	    Logger.log("chan_mgr", "warning: no processor defined...");
	    return;
	}
	
	try {
	    processor_.process(msg);
	}
	catch (Exception e) {
	    Logger.log("chan_mgr", "error: message processing failed " + e);
	}
    }

}
