/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/


package casa.dodwan.net;

import java.io.*;
import java.net.*;
import java.net.InetSocketAddress;
import java.util.Set;

import casa.dodwan.util.ArgumentNotFoundException;
import casa.dodwan.util.ArgumentParsing;

import casa.util.Logger;
import casa.dodwan.message.Message;
import casa.dodwan.util.SystemEnvironment;
import casa.dodwan.util.DataFormatting;

public class TCP_Server
    extends Thread implements Channel {

    private int port_;
    private ServerSocket socket_;
    private ChannelManager manager_;
    private boolean isOpen_ = false;
    private int bufSize_ = 0;

    private long lastRx_  = System.currentTimeMillis();
    
    // ------------------------------------------------------------

    public static String getSyntax() {

	return "[-lp localPort] [-bs bufSize]";
    }

    // ------------------------------------------------------------
    public TCP_Server(String[] arguments,
		      ChannelManager manager)
	throws Exception {

	int localPort = 0;
	int bufSize   = 0;
	
	try {
	    localPort = ArgumentParsing.searchInt("-lp", arguments);
	} catch (ArgumentNotFoundException e) {}
	try {
	    bufSize = ArgumentParsing.searchInt("-bs", arguments);
	}
	catch (ArgumentNotFoundException e) {}

	init(localPort, manager, bufSize);
    }

    // ------------------------------------------------------------
    public TCP_Server(int port,
		      ChannelManager manager,
		      int bufSize)
	throws Exception {

	init(port, manager, bufSize);
    }

    // ------------------------------------------------------------
    public void init(int port,
		     ChannelManager manager,
		     int bufSize)
	throws Exception {
	
	setDaemon(true);
	   
	port_    = port;
	manager_ = manager;
	bufSize_ = bufSize;
	
	socket_ = new ServerSocket(port_);
	if (bufSize_ > 0)
	    socket_.setReceiveBufferSize(bufSize_);
	
	Logger.log("tcp_server", "init " + this);
    }

    // ------------------------------------------------------------
    public int getPort() {
	return socket_.getLocalPort();
    }
    
    // ------------------------------------------------------------
    public void open() {

	if (isOpen_)
	    return;

	this.start();
	isOpen_ = true;
    }

    // ------------------------------------------------------------
    public void close() {

	Logger.log("tcp_server", "close " + this);

	try {
	    socket_.close();
	}
	catch (Exception e) {}
	isOpen_ = false;
    }

    // ------------------------------------------------------------
    public boolean isOpen() {
	return isOpen_;
    }

    // ------------------------------------------------------------
    public Set<String> getPeers() {
	return null;
    }

    // ------------------------------------------------------------
    public void write(Message msg) {}

    // ------------------------------------------------------------
    public void run() {
	
	try {
	    while(! socket_.isClosed()) {
		Socket peerSocket = socket_.accept();
		if (bufSize_ > 0) {
		    peerSocket.setSendBufferSize(bufSize_);
		    peerSocket.setReceiveBufferSize(bufSize_);
		}

		lastRx_ = System.currentTimeMillis();
		
		Logger.log("tcp_server", "cx_accept (channel=" 
		 	   + peerSocket + ")");

		TCP_Channel channel = 
		    new TCP_Channel(peerSocket, manager_);
		
		if (manager_ != null) {
		    manager_.addChannel(channel);
		}
		channel.start();
	    }
	    // serverSocket.close(); // Never reached
	}
	catch (Exception e) {
	    Logger.log("tcp_server", "socket closed");
	    // e.printStackTrace();
	}
    }

    // ------------------------------------------------------------
    public String toString() {

	if ((socket_ == null) || socket_.isClosed())
	    return "TCP_Server()";

	String buf = null;
	try {
	    buf = ",buf=" + socket_.getReceiveBufferSize();
	}
	catch (Exception e) {}
	
	return "TCP_Server(local="
	    + DataFormatting.toString(socket_.getInetAddress())
	    + ":"
	    + socket_.getLocalPort()
	    + (buf!=null?buf:"")
	    + ")";
    }
    
    // ------------------------------------------------------------
    public long lastRx() {
	return lastRx_;
    }
   
}
