/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.run;

import java.util.Random;
import java.util.Date;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.FileReader;
import java.io.PrintStream;
import java.io.File;
import java.lang.Runtime;

import casa.dodwan.config.DodwanEnvironment;
import casa.dodwan.util.ArgumentParsing;
import casa.dodwan.util.Command;
import casa.dodwan.util.SystemEnvironment;
import casa.dodwan.util.Console;
import casa.dodwan.util.TCPConsole;
import casa.dodwan.util.RandomSeed;
import casa.dodwan.util.DataFormatting;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import casa.util.Logger;

// FG ---
// import casa.boson.*;
// import casa.boson.scenarios.*;
// import casa.dodwan.app.DodwanNode;
// --- FG

public class dodwand extends Thread {

    // This class defines a wrapper for an instance of class Dodwan.

    
    // ==================================================
    private class TerminationThread
	extends Thread {

	private long wait_;
	private dodwand wrapper_;
	
	public TerminationThread(long wait, dodwand wrapper) {
	    wait_ = wait;
	    wrapper_ = wrapper;
	}
	
	public void run() {
	    try {
		if (wait_ <= 0)
		    return;

		Logger.log("dodwan", "will terminate in " + wait_ + " ms");
		
		sleep(wait_);
		
		System.out.println("terminating as planned");
		Logger.log("dodwan", "shutting down (TerminationThread)");

		// Calling System.exit(), so the ShutdownHook below
		// will be executed
		System.exit(0);
		
	    } catch (Exception e) {
		Logger.log("dodwan", "TerminationThread: " + e);
	    }
	}
    }
    // ==================================================
    private class ShutdownHook
	extends Thread {

	private dodwand wrapper_;
	
	public ShutdownHook(dodwand wrapper) {
	    wrapper_ = wrapper;
	}
	
	public void run() {

	    try {
		System.out.println("Got a termination signal");

		Logger.log("dodwan", "shutting down (ShutdownHook)");
		wrapper_.dodwan.offline();
		Logger.log("dodwan", "SH1");
                wrapper_.deactivatePlugins();
		Logger.log("dodwan", "SH2");
		wrapper_.clear();
		Logger.log("dodwan", "SH3");
		wrapper_.dodwan.doNotify(); 
		Logger.log("dodwan", "SH4");
	    } catch (Exception e) {
		Logger.log("dodwan", "ShutdownHook: " + e);
	    }
	}
    }

    // ==================================================

    public Dodwan dodwan   = null;
    private String ifile    = null;
    private String cfile    = null;
    private File portFile = null;
    private PrintWriter portWriter = null;
    private long begin      = -1;
    private long end        = -1;
    private long seed       = -1;
    private Console console = null;
    private File pid_file   = null;
    private boolean cleared = false;
    private List<ReflectedPlugin> rplugins = new ArrayList<ReflectedPlugin>();

    // FG ---
    // private DisseminationLogsScenario app = null;
    // --- FG

    // ------------------------------------------------------------
    public dodwand(String args[]) throws Exception {

	// casa.dodwan.util.Catalog.main(args);
	// System.exit(1);
	
	if (ArgumentParsing.searchIndex("-h", args) != -1) {
	    exit(0);
	}
	
	begin = -1;
	try {
	    begin = ArgumentParsing.searchLong("-begin", args);
	} catch (Exception e) {};
	    
	end = -1;
	try {
	    end = ArgumentParsing.searchLong("-end", args);
	} catch (Exception e) {};
	
	seed = System.currentTimeMillis();
	try {
		seed = ArgumentParsing.searchLong("-seed", args);
	} catch (Exception e) {};
	RandomSeed.setSeed(seed);
	
	cfile = null;
	try {
	    cfile = ArgumentParsing.searchString("-c", args);
	} catch (Exception e) {};

	ifile = null;
	try {
	    ifile = ArgumentParsing.searchString("-i", args);
	} catch (Exception e) {};

	
	DodwanEnvironment env = DodwanEnvironment.getInstance();
	env.displayProperties();
	pid_file = new File(env.directory, "pid");
	
	System.out.println("Instantiation of Dodwan on " + SystemEnvironment.host);
	
	dodwan = new Dodwan();

	// FG ---
	String appName = null;
	try {
	    appName = ArgumentParsing.searchString("-app", args);
	} catch (Exception e) {};
	if (appName != null) {
	    // SimulNode node = new DodwanNode(dodwan, SystemEnvironment.host);
	    // app = new DisseminationLogsScenario();
	    // app.setSimulNode(node);
	    // app.setApplicationScenario(appName, begin);
	}
	// --- FG
	
	console = new Console();
	console.addSubConsole("d", "[d]odwan", dodwan.console());
	
	int console_port = env.console_port;
	if (console_port != -1) {
	    TCPConsole mainConsole =
		new TCPConsole(console, console_port);
	    if (console_port == 0) {
		console_port = mainConsole.getLocalPort();
                env.setConsolePort(console_port);
            }
	    System.out.println("A console is now listening to port TCP#" + console_port);
	}

        
        openPortFile();
        writePortFile("dodwan.console_port", console_port);
        
        // Instantiation and starting of each plugin
        activatePlugins();
         
        closePortFile();
        
        
        // Execution of console command file 
	if (cfile != null) {
	    System.out.println("Parsing command file " + cfile);
	    console.parseCommandFile(cfile, System.out);
	}
    }

    
    
  
        
    // ------------------------------------------------------------
    private void openPortFile() throws FileNotFoundException {
        this.portFile = new File(DodwanEnvironment.getInstance().directory, "ports");
        this.portWriter = new PrintWriter(new FileOutputStream(this.portFile));
    }
    
    // ------------------------------------------------------------
    public synchronized void writePortFile(String prop, int port) {
        this.portWriter.println(prop + "=" + port);
    }
            
    // ------------------------------------------------------------
    private void closePortFile() {
        this.portWriter.close();
        System.out.println("File " + this.portFile + " created");
    }
            
    // ------------------------------------------------------------
    private void deletePortFile() {
        if (this.portFile != null) {
             this.portFile.delete();
             System.out.println("File " + this.portFile + " deleted");
             this.portFile = null;
        }
    }
    
    
    
    // ------------------------------------------------------------
    private void activatePlugins() {
        final String [] plugins = DodwanEnvironment.getInstance().properties.getString("dodwan.plugins", "").split(",");
        for (String elt:plugins) {
            
           String plugin = elt.trim();

           if (!plugin.equals("")) {
            
            ReflectedPlugin rplugin = new ReflectedPlugin();
            
            rplugin.name = plugin;
            
            // Determine the name of the class corresponding to the name of the plugin
            // example dodwan-napi-ws -> DodwanNapiWs
            StringBuffer pluginClassName = new StringBuffer();
            pluginClassName.append(Character.toUpperCase(plugin.charAt(0)));
            char cc;
            int i=1;
            while (i < plugin.length()) {
               cc = plugin.charAt(i);
               if (cc == '-') {
                 i++;
                 if (i < plugin.length()) cc = Character.toUpperCase(plugin.charAt(i));
               }
               pluginClassName.append(cc); 
               i++;
            }
 
            try {
                System.out.println("Launching plugin "+plugin+" (class "+pluginClassName+")");
                rplugin.pluginClass = Class.forName("casa.dodwan.plugin." + pluginClassName.toString());
                rplugin.pluginConstructor = rplugin.pluginClass.getConstructor(new Class [] {});
                rplugin.pluginObject = rplugin.pluginConstructor.newInstance(null);
                rplugin.startMethod = rplugin.pluginClass.getMethod("startPlugin", new Class [] {dodwand.class});
                rplugin.stopMethod = rplugin.pluginClass.getMethod("stopPlugin", new Class [] {});
                Object ret = rplugin.startMethod.invoke(rplugin.pluginObject, new Object [] {this});   
            }
            catch (ClassNotFoundException ex) {
               System.err.println("Plugin class not found for plugin " + plugin);
               ex.printStackTrace(System.err);
            }
            catch (NoSuchMethodException ex ) { 
               System.err.println("Cannot find constructor or method startPlugin for plugin " + plugin); 
               ex.printStackTrace(System.err);
            }
            catch (InstantiationException ex ) { 
                System.err.println("Cannot instantiate plugin class " + plugin); 
                ex.printStackTrace(System.err);
            } 
            catch (IllegalAccessException ex ) { 
                System.err.println("No permission to call methods of plugin class " + plugin);
                ex.printStackTrace(System.err);
            } 
            catch (InvocationTargetException ex) {
                System.err.println("Exception thrown by constructor or method startPlugin for plugin " + plugin);
                ex.printStackTrace(System.err);
            }
                 
            this.rplugins.add(rplugin);
          }
        }
    }
    
    
    
    
    // ------------------------------------------------------------
    private void deactivatePlugins() {
        for (ReflectedPlugin rplugin : this.rplugins) {
            try {
                rplugin.stopMethod.invoke(rplugin.pluginObject);
            } catch (IllegalAccessException ex) {
                System.err.println("No permission to call method stopPlugin for plugin " + rplugin.name);
                ex.printStackTrace(System.err);
            } catch (IllegalArgumentException ex) {
                System.err.println("Argument error when calling method stopPlugin for plugin for plugin " + rplugin.name);
                ex.printStackTrace(System.err);
            } catch (InvocationTargetException ex) {
                System.err.println("Exception thrown by method stopPlugin for plugin " + rplugin.name);
                ex.printStackTrace(System.err);
            }
        }
    }
    
    
    
    // ------------------------------------------------------------
    public void run() {
	
	Runtime.getRuntime().addShutdownHook(new ShutdownHook(this));

	if (begin != -1) {
	    System.out.println("Deferring start until EPOCH " + begin
			       + " ms (" + new Date(begin) + ")");
	    long now = System.currentTimeMillis();
	    long wait = begin - now;
	    if (wait > 0)
		try {
		    Thread.sleep(wait);
		}
		catch (Exception e) {}
	}
	
	long wait = -1;
	if (end != -1) {
	    long now = System.currentTimeMillis();
	    wait = end - now;
	    System.out.println("Will stop in " +
			       DataFormatting.to_hh_mm_ss(wait));
	    TerminationThread th = new TerminationThread(wait, this);
	    th.start();
	}

	// FG ---
	// if (app != null)
	//     app.start();
	    // new Thread(app).start();
	// --- FG
	
	// Switch DoDWAN online
	dodwan.online();

	// If option '-i' is specified, start a console, taking as
	// input either stdin or a FIFO file
	boolean interactive = (ifile != null);
	if (interactive) {
	    System.out.println("Input from " + ifile);
	    PrintStream out = System.out;
	    boolean fifo = (! ifile.equals("-"));	
	    try {
		do {
		    BufferedReader in = (fifo?
					 new BufferedReader(new FileReader(ifile)):
					 new BufferedReader(new InputStreamReader(System.in)));
		    console.parseCommandStream(in, out, true);
		} while (fifo);
	    }
	    catch (Exception e) {
		System.out.println("Failed to open " + ifile);
	    }
	}
    }

    // ------------------------------------------------------------
    public void clear() {

	if (cleared)
	    return;
	
	// Trying to remove pid file (assuming there is one)
	if (pid_file.exists()) {
	    try {
		pid_file.delete();
		System.out.println("File " + pid_file + " deleted");
		pid_file = null;
	    }
	    catch (Exception e) {
		e.printStackTrace();
	    }
	}
	
	// Trying to remove port file (assuming there is one)
	deletePortFile();

	cleared = true;
    }
    
    // ------------------------------------------------------------
    public static void exit(int error) {
	help();
	Command.exit(error);
    }

    // ------------------------------------------------------------
    public static void help() {
	System.out.println("\n"
			   + "Syntax: dodwand [-h] [-c file] [-i (-|fifo file)] [-seed value] [-begin time (ms)] [-end time (ms)] -Dkey=value ...");
	System.out.println("  -h: help");
	System.out.println("  -c: command file (read once at startup)");
	System.out.println("  -i: interactive mode, with a console taking either stdin or a fifo file as input (default: no console)");
	System.out.println("  -seed:  seed for the random generator");
	System.out.println("  -begin: deferred start time (EPOCH in ms)");
	System.out.println("  -end:   termination time (EPOCH in ms)");
    }

    // ------------------------------------------------------------
    public static void main(String[] args) throws Exception {

	dodwand wrapper = new dodwand(args);
	wrapper.start();

	// Wait until Dodwan is shut down (either via a 'do shutdown'
	// issued in a console, or via the builtin TerminationThread
	// if end time is specified)
	wrapper.dodwan.doWait();
	
	// Try to terminate DoDWAN gracefully
	wrapper.clear();
	System.out.println("DoDWAN terminated");
	System.out.flush();
	
    }

    
    // =============================================================
    
    // Product of the reflection API for a plugin
    private class ReflectedPlugin {
        String name;
        Class pluginClass ;
        Constructor pluginConstructor;
        Method startMethod;
        Method stopMethod;
        Object pluginObject;
    }
    
}

