/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.sendreceive;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import casa.dodwan.config.DodwanEnvironment;
import casa.dodwan.docware.Descriptor;
import casa.dodwan.message.Message;
import casa.dodwan.util.Console;
import casa.util.Processor;
import casa.dodwan.util.Service;
import casa.dodwan.util.Sink;

public class SendReceiveService 
    extends Service
    implements Sink<Message> {
    
    private class MessageReceiver {

	Descriptor pattern;

	Processor<Message> processor;

	public MessageReceiver(Descriptor pattern,
			       Processor<Message> processor) {
	    this.pattern = pattern;
	    this.processor = processor;
	}
    }

//      private Dispatcher<Message> loop_ = new Dispatcher<Message>();

//      private Dispatcher<Message> receivers_ = new Dispatcher<Message>();
    
    private Map<String, MessageReceiver> map_ = 
	new HashMap<String, MessageReceiver>();

    private Sink<Message> sink_ = null;

    // ------------------------------------------------------------
    public SendReceiveService(Sink<Message> sink)
	throws Exception {
	
	super.suspend();
	
	sink_ = sink;
//  	verbosity.enable();
    }
    
    // ------------------------------------------------------------
    public void addSubscription(String key, Descriptor pattern, 
				Processor<Message> processor) {
	
	if (map_.containsKey(key)){
	    System.out.println("The key " + key + ", is already registered.");
	    return;
	}

	if (verbosity.isEnabled())
	    System.out.println("SendReceiveService.addSubscriber(" 
			       + key + ", "
			       + pattern + ", "
			       + processor + ")");

	MessageReceiver receiver = 
	    new MessageReceiver(pattern, processor);
	map_.put(key, receiver);
    }

    // ------------------------------------------------------------
    public void removeSubscription(String key) {
	
	if (verbosity.isEnabled())
	    System.out.println("SendReceiveService.removeSubscription("
			       + key + ")");
	
	map_.remove(key);
    }
    
    // ------------------------------------------------------------
    public void send(Message message) throws Exception {
	
	if (verbosity.isEnabled())
	    System.out.println("SendReceiveService.send()");
	
	 if (isSuspended()) {
	     if (verbosity.isEnabled())
		 System.out.println("  This service is suspended. Aborting...");
	     return;
	 }

	 int max_payload = DodwanEnvironment.getInstance().max_payload_size;
	 if ((max_payload > 0) 
	     && (message.getPayload().length > max_payload)) {

	     System.out.println("SendReceiveService.send(msg): discarding message because its payload is too long ("
				+ message.getPayload().length 
				+ " > max="
				+ max_payload 
				+ ")");
	     return;
	 }
				
	 casa.util.Logger.log("sendreceive", 
				     "> id=" 
				     + message.getDescriptor().getDocumentId() + ")");
	
	 sink_.write(message);
    }

    // ------------------------------------------------------------
    public Set<String> getKeys() {
	
	return map_.keySet();
    }
    
    // ------------------------------------------------------------
    public Console console() {
//  	if (console_ == null)
//  	    console_ = new PubSubServiceConsole(this);
//  	return console_;
	return null;
    }
    
    // ------------------------------------------------------------
    public void resume() {
	
	if (verbosity.isEnabled())
	    System.out.println("SendReceiveService.resume()");
	
	if (isEnabled()) {
	    super.resume();
	}
	else if (verbosity.isEnabled())
	    System.out.println("  Failure: this service is disabled");
    }
    
    // ------------------------------------------------------------
    public void setVerbosity(int level) {
	
	switch (level) {
	case 0:
	    verbosity.disable();
	    break;
	default:
	    verbosity.enable();
	    break;
	}
	
	System.out.println("SendReceiveService.setVerbosity(" + level + ")");
    }
    
    // ------------------------------------------------------------
    @Override
	public void suspend() {
	
	if (verbosity.isEnabled())
	    System.out.println("SendReceiveService.suspend()");
	
	if (isEnabled()) {
	    super.suspend();
	}
	else if (verbosity.isEnabled())
	    System.out.println("  Failure: this service is disabled");
    }
    
    // ------------------------------------------------------------
    public void write(Message message)
	throws Exception {

	Descriptor desc = message.getDescriptor();

	if (verbosity.isEnabled())
	    System.out.println("SendReceiveService.write('"
			       + desc.getDocumentId() + "')");
	
	if (isSuspended()) {
	    if (verbosity.isEnabled())
		System.out.println("  This service is suspended. Aborting...");
	    return;
	}
	
	String serviceType = desc.getAttribute("_stype");

//  	System.out.println("SendReceiveService.write()");
	for (MessageReceiver recv: map_.values()) {
	    
	    boolean match = desc.matches(recv.pattern);
//  	    System.out.println("  Message matches pattern: " + match 
//  			       + (match?" " + recv.pattern:""));

	    if (match) 
		recv.processor.process(message);
	}
    }

    // ------------------------------------------------------------
    
}
