/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.transmission;

import java.io.BufferedReader;
import java.io.PrintStream;
import java.net.InetAddress;
import java.net.Inet4Address;
import java.net.Inet6Address;
import java.net.InetSocketAddress;
import java.net.NetworkInterface;
import java.util.Map;
import java.util.Enumeration;
import java.lang.reflect.Method;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.Throwable;

import casa.dodwan.message.Message;
import casa.dodwan.message.MessageBufferizer;
// import casa.dodwan.net.InetSocket;
// import casa.dodwan.net.MulticastInetSocket;
import casa.dodwan.util.ArgumentNotFoundException;
import casa.dodwan.util.ArgumentParsing;
import casa.dodwan.util.ArgumentParsingException;
import casa.dodwan.util.DataFormatting;
import casa.dodwan.util.Command;
import casa.dodwan.util.Console;
import casa.dodwan.util.Dispatcher;
import casa.dodwan.util.Socket;
import casa.dodwan.config.DodwanEnvironment;
import casa.dodwan.util.Environment;
import casa.dodwan.net.Channel;
import casa.dodwan.net.ChannelManager;
import casa.dodwan.net.TCP_Server;

public class TransmissionServiceConsole extends Console {
    
    public TransmissionService transmissionService_;
    
    // ------------------------------------------------------------
    public TransmissionServiceConsole(String prompt,
				      TransmissionService transmissionService) {
	
	transmissionService_ = transmissionService;
	prompt_ = prompt;
    }
    
    // ============================================================
    // MAIN COMMANDS
    // ============================================================
    
    // ------------------------------------------------------------
    public TransmissionServiceConsole(TransmissionService transmissionService) {
	
	this("trans% ", transmissionService);
    }
    
    // --------------------------------------------------------------------
    public void help(String offset, PrintStream out) {
	out.println(offset   + "[a]dd channelId [-t className] args...\n"
		    + offset + "[e]nable / [d]isable\n"
		    + offset + "[j]oin [-a mcastaddr] <-i if>\n"
		    + offset + "[le]ave [-a mcastaddr] <-i if>\n"
		    + offset + "[li]st\n"
		    + offset + "[n]eighbourhood\n"
		    // + offset + "[set] <-tx dropRatio> <-rx dropRatio> <-ft fragmentationThreshold> \n"
		    + offset + "[rem]ove channelId|serverId\n"
		    + offset + "[su]spend / [res]ume\n" 
		    + offset + "[st]atus\n" 
		    + offset + "[setf] <-ft fragmentationThreshold> \n"
		    + offset + "[sen]d <file+>\n"
		    + offset + "[ser]ver <serverId> <port>\n"
		    + offset + "[rec]eive <file+>\n"
		    + offset + "[v]erbosity <level>\n");
    }
    
    // ============================================================
    @Override
	public void processCommand(String[] arguments, BufferedReader reader,
				   PrintStream out) {
	
	if (arguments.length < 1) {
	    help("  ", out);
	    return;
	}
	
	if (arguments[0].startsWith("a")) {
	     transmissionAddChannel(arguments, out);
	     return;
	}
	
	if (arguments[0].startsWith("e")) {
	    transmissionEnable(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("d")) {
	    transmissionDisable(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("frag")) {
	    transmissionFragmentation(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("j")) {
	    transmissionJoin(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("l")) {
	    transmissionList(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("n")) {
	    transmissionNeighbourhood(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("rec")) {
	    transmissionReceive(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("rem")) {
	    transmissionRemove(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("res")) {
	    transmissionResume(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("sen")) {
	    transmissionSend(arguments, out);
	    return;
	}

	if (arguments[0].startsWith("ser")) {
	    transmissionAddServer(arguments, out);
	    return;
	}
	    
	if (arguments[0].startsWith("st")) {
	    transmissionStatus(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("su")) {
	    transmissionSuspend(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("v")) {
	    transmissionVerbosity(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("h")) {
	    out.println("Transmission Service: Command Set\n");
	    help("  ", out);
	    return;
	}
	
	out.println("Unknown command");
	help("  ", out);
    }
    
    // ------------------------------------------------------------
    public void resume(String[] arguments, PrintStream out) {
	
	if (arguments.length != 0) {
	    help("  ", out);
	    return;
	}
	
	transmissionService_.resume();
    }
    
    // ------------------------------------------------------------
    public void suspend(String[] arguments, PrintStream out) {
	
	if (arguments.length != 0) {
	    help("  ", out);
	    return;
	}
	
	transmissionService_.suspend();
    }
    
    // ------------------------------------------------------------
    public void transmissionDisable(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	transmissionService_.disable();
    }
    
    // ------------------------------------------------------------
    public void transmissionEnable(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	transmissionService_.enable();
    }

    
    // ------------------------------------------------------------
    public void transmissionAddServer(String[] arguments,
				      PrintStream out) {
	
	ChannelManager channelMgr = transmissionService_.channelManager;

	if (arguments.length <= 2) {
	    help("  ", out);
	    return;
	}

	String id = arguments[1];
	if (id.startsWith("-")) {
	    out.println("Invalid id");
	    return;
	}

	if (channelMgr.hasChannel(id)) {
	    out.println("A channel is already registered with this id");
	    return;
	}
	    
	if (channelMgr.hasServer(id)) {
	    out.println("A server is already registered with this id");
	    return;
	}

	try {
	    int port = Integer.parseInt(arguments[2]);
	    int bufSize = 0;
	    if (arguments.length > 3)
		bufSize = Integer.parseInt(arguments[3]);
	    
	    TCP_Server server =
		new TCP_Server(port,
			       transmissionService_.channelManager, bufSize);
	    server.start();
	    channelMgr.addServer(id, server);
	}
	catch (Exception e) {
	    out.println("Could not start TCP_Server on port " + arguments[2]);
	}
    }
    
   // ------------------------------------------------------------
    public void transmissionAddChannel(String[] arguments,
				       PrintStream out) {

	if (arguments.length <= 2) {
	    help("  ", out);
	    return;
	}
	
	DodwanEnvironment env = DodwanEnvironment.getInstance();
	ChannelManager channelMgr = transmissionService_.channelManager;
	
	String id = arguments[1];
	if (id.startsWith("-")) {
	    out.println("Invalid id");
	    return;
	}

	if (channelMgr.hasChannel(id)) {
	    out.println("A channel is already registered with this id");
	    return;
	}

	if (channelMgr.hasServer(id)) {
	    out.println("A server is already registered with this id");
	    return;
	}

	String className = null;
    	try {
	    className = ArgumentParsing.searchString("-t", arguments);
    	} catch (Exception e) {
    	    out.println(e);
	    return;
    	}

	Channel channel = null;

	if (className.equals("udp")) 
	    className = "casa.dodwan.net.UDP_Channel";
	else if (className.equals("tcp")) 
	    className = "casa.dodwan.net.TCP_Channel";

	try {
	    channel = getChannel(className, arguments, out, channelMgr);
	    
	    channelMgr.addChannel(id, channel);
	    
	    if (! transmissionService_.isSuspended())
		channel.open();
	} catch (Exception e) {
	    out.println("Channel creation failed (" + e + ")");
    	}
    }

    // ------------------------------------------------------------
    private Channel getChannel(String className,
			       String[] arguments,
			       PrintStream out,
			       ChannelManager channelManager)
	throws Exception {

	Channel channel = null;
	try {
	    Class class_ = Class.forName(className);
	    // Method method_ = class_.getMethod("getInstance", String[].class);
	    // channel = (Channel)method_.invoke(null, (Object)arguments);
	    Constructor constructor =
		class_.getConstructor(new Class[]{String[].class,ChannelManager.class});
	    channel = (Channel)constructor.newInstance((Object)arguments,
						       channelManager);
	}
	catch (InvocationTargetException e) {
	    Throwable cause = e.getCause();
	    throw new Exception(cause.getMessage());
	}
	catch (Exception e) {
	    throw e;
	}
	
	return channel;
    }

    // ------------------------------------------------------------
    public void transmissionRemove(String[] arguments, PrintStream out) {

	if (arguments.length <= 1) {
	    help("  ", out);
	    return;
	}
	
	ChannelManager channelMgr = transmissionService_.channelManager;
	
	String id = arguments[1];
	
	if (channelMgr.hasChannel(id)) {
	    channelMgr.removeChannel(id);
	    return;
	}

	if (channelMgr.hasServer(id)) {
	    channelMgr.removeServer(id);
	    return;
	}
	    
	out.println("This id is not registered");
    }

    // ------------------------------------------------------------
    public void transmissionList(String[] arguments, PrintStream out) {

	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}

	ChannelManager channelMgr = transmissionService_.channelManager;

	out.println();
	out.println("Servers");
	out.println(DataFormatting.toString("Id", 10)
		    + "\tServer");
	
	Map<String,TCP_Server> servers = channelMgr.getServers();
	for (String id: servers.keySet())
	    out.println(DataFormatting.toString(id, 10)
			+ "\t" + servers.get(id));
	out.println();
	
	out.println("Channels");
	out.println(DataFormatting.toString("Id", 10)
		    + "\tChannel");

	Map<String,Channel> channels = channelMgr.getChannels();
	for (String id: channels.keySet())
	    out.println(DataFormatting.toString(id, 10)
			+ "\t" + channels.get(id));
	out.println();
    }
    
    // ------------------------------------------------------------
    public void transmissionJoin(String[] arguments, PrintStream out) {
	
	// FG: a revoir
	// MulticastInetSocket socket = null;
	try {
	    // socket = (MulticastInetSocket) (transmissionService_.msgSocket.socket);
	} catch (Exception e) {
	    out.println("The socket is not a MulticastInetSocket.");
	    return;
	}
	
	String mcastAddress = null;
	String infAddress = null;
	
	try {
	    try {
		mcastAddress = ArgumentParsing.searchString("-a", arguments);
	    } catch (ArgumentNotFoundException e) {} 
	    try {
		infAddress = ArgumentParsing.searchString("-i", arguments);
	    } catch (ArgumentNotFoundException e) {} 
	}
	catch (ArgumentParsingException e) {
	    Command.parsingException(e);
	    return;
	} catch (Exception e) {
	    out.println(e);
	}
	
	if (mcastAddress == null) {
	    try {
		out.println("Attempting to reset former multicast subscription");
		// socket.joinAgain();
		return;
	    }
	    catch (Exception e) {
		out.println("   Failed");
	    }
	}

	if (infAddress == null) {
	    try {
		InetAddress address = InetAddress.getByName(mcastAddress);
		out.println("Attempting to join group " 
			    + address);
		// socket.joinGroup(address);
	    } catch (Exception e) {
		out.println("   Failed");
	    }
	} else {
	    try {
		InetAddress address = InetAddress.getByName(mcastAddress);
		InetAddress netInf = InetAddress.getByName(infAddress);
		out.println("Attempting to join group " 
			    + address 
			    + " on interface "
			    + netInf);
		// socket.joinGroup(address, netInf);
	    } catch (Exception e) {
		out.println("   Failed");
	    }
	}
    }
    
    // ------------------------------------------------------------
    public void transmissionLeave(String[] arguments, PrintStream out) {
	
	// FG: a revoir
	// MulticastInetSocket socket = null;
	try {
	    // socket = (MulticastInetSocket) (transmissionService_.msgSocket.socket);
	} catch (Exception e) {
	    out.println("The socket is not a MulticastInetSocket.");
	    return;
	}
	
	String mcastAddress = null;
	String infAddress = null;
	
	try {
	    mcastAddress = ArgumentParsing.searchString("-a", arguments);
	    
	    try {
		infAddress = ArgumentParsing.searchString("-i", arguments);
	    } catch (ArgumentNotFoundException e) {
	    }
	} catch (ArgumentNotFoundException e) {
	    Command.missingArgument(e);
	    return;
	} catch (ArgumentParsingException e) {
	    Command.parsingException(e);
	    return;
	} catch (Exception e) {
	    out.println(e);
	}
	
	if (infAddress == null) {
	    try {
		InetAddress address = InetAddress.getByName(mcastAddress);
		out.print("Attempting to leave group " + address);
		// socket.leaveGroup(address);
	    } catch (Exception e) {
		out.println("   Failed");
	    }
	} else {
	    try {
		InetSocketAddress address = new InetSocketAddress(mcastAddress,
								  0);
		NetworkInterface netInf = NetworkInterface
		    .getByInetAddress(InetAddress.getByName(infAddress));
		out.print("Attempting to leave group " 
			  + address
			  + " on interface "
			  + netInf);
		// socket.leaveGroup(address, netInf);
	    } catch (Exception e) {
		out.println("   Failed");
	    }
	}
    }
    
    // ------------------------------------------------------------
    public void transmissionReceive(String[] arguments, PrintStream out) {
	
	if (arguments.length <= 1) {
	    help("  ", out);
	    return;
	}
	
	Dispatcher<Message> dispatcher = transmissionService_.dispatcher;
	
	for (int i = 1; i < arguments.length; i++) {
	    try {
		String name = arguments[i];
		out
		    .println("Emulating the receipt of the document stored in file "
			     + name);
		Message doc = new MessageBufferizer().fromFile(name);
		dispatcher.write(doc);
	    } catch (Exception e) {
		out.println(e);
	    }
	}
    }
    
    // ------------------------------------------------------------
    public void transmissionResume(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	transmissionService_.resume();
    }
    
    // ------------------------------------------------------------
    public void transmissionSend(String[] arguments, PrintStream out) {
	
	if (arguments.length <= 1) {
	    help("  ", out);
	    return;
	}

	// FG: a revoir
	// InetSocket<Message> socket = transmissionService_.msgSocket.socket;
	
	for (int i = 1; i < arguments.length; i++) {
	    try {
		String name = arguments[i];
		out.println("Sending the document stored in file " + name);
		Message doc = new MessageBufferizer().fromFile(name);
		out.println("Sorry: cannot do that right now. Check the code...");
		// socket.write(doc);
	    } catch (Exception e) {
		out.println(e);
	    }
	}
    }
    
    // ------------------------------------------------------------
    public void transmissionFragmentation(String[] arguments, PrintStream out) {
	
	try {
	    try {
		int threshold = ArgumentParsing.searchInt("-ft", arguments);
		DodwanEnvironment.getInstance().fragmentation_threshold = threshold;
	    } catch (ArgumentNotFoundException e) {}
	    
	} catch (ArgumentNotFoundException e) {
	    Command.missingArgument(e);
	    return;
	} catch (ArgumentParsingException e) {
	    Command.parsingException(e);
	    return;
	} catch (Exception e) {
	    out.println(e);
	}
    }
    
    // ------------------------------------------------------------
    public void transmissionStatus(String[] arguments, PrintStream out)	{

	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	out.println(transmissionService_);

	try {
	    for (Enumeration e = NetworkInterface.getNetworkInterfaces() ; e.hasMoreElements() ;) {
		NetworkInterface iface = (NetworkInterface)e.nextElement();
		out.println("\nName : " + iface.getDisplayName()
				   + (iface.isUp()?" (UP)":" (DOWN)")
				   + (iface.supportsMulticast()?" (MCAST)":""));
		for (Enumeration<InetAddress> a = iface.getInetAddresses(); a.hasMoreElements();) {
		    InetAddress add = a.nextElement();
		    out.println((add instanceof Inet6Address?"Inet6: ":"Inet4: ")
				       + add);
		}
	    }
	}
	catch (Exception e) {}
    }
    
    // ------------------------------------------------------------
    public void transmissionNeighbourhood(String[] arguments, PrintStream out) {

	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	out.println("Address\t\t\tId\n");
	Environment env = transmissionService_.environment;

	out.println("Code à revoir !");
	// for (String id: env.getIds())
	//     // FG: a revoir
	//     out.println(transmissionService_.msgSocket.defaultChannel.remoteAddress(id) + "\t" + id);
}
    
    // ------------------------------------------------------------
    public void transmissionSuspend(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	transmissionService_.suspend();
    }
    
    // ------------------------------------------------------------
    public void transmissionVerbosity(String[] arguments, PrintStream out) {
	
	if (arguments.length < 2) {
	    help("  ", out);
	    return;
	}
	
	int level = -1;
	try {
	    level = Integer.parseInt(arguments[1]);
	} catch (Exception e) {
	    out.println("Could not parse verbosity level");
	}
	
	if (level >= 0)
	    transmissionService_.setVerbosity(level);
    }
}
