/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.util;

import java.util.Collection;
import java.util.HashSet;
import java.io.File;


public class FileBasedRepository<OBJ_TYPE> 
    implements Repository<String, OBJ_TYPE> {
    
    private File path_ = null;
    private Bufferizer<OBJ_TYPE> bufferizer_ = null;
    
    public Verbosity verbosity = new Verbosity();
    
    // ------------------------------------------------------------
    public FileBasedRepository(File path,
			       Bufferizer<OBJ_TYPE> bufferizer) {

// 	System.out.println("FileBasedRepository("
// 			   + path + ")");

	path_ = path;

	if (!path.exists())
	    path.mkdirs();
	
	bufferizer_ = bufferizer;

// 	verbosity.enable();
    }

    // ------------------------------------------------------------
    public void clear() {

	File[] files = path_.listFiles();

	for (File file: files) {
	    file.delete();
	}
    }
    
    // ------------------------------------------------------------
    public boolean contains(String key) {

	File file = new File(path_, key);
	return file.exists();
    }
    
    // ------------------------------------------------------------
    public void delete(String key) {
	
	if (verbosity.isEnabled())
	    System.out.println("FileBasedRepository.delete(" + key + ")");
	
	    try {
		File file = new File(path_, key);
		file.delete();
	    }
	    catch (Exception e) {
		System.out.println("FileBasedRepository.delete(): failed");
	    }
    }
    
    // ------------------------------------------------------------
    public void put(String key, OBJ_TYPE obj) {
	
	if (verbosity.isEnabled())
	    System.out.println("FileBasedRepository.put(" + key + ")");

	try {
//  	    System.out.println("Trying to put document "
//  			       + key + " in " + path_);
	    bufferizer_.toFile(obj, new File(path_, key));
	}
	catch (Exception e) {
	    e.printStackTrace();
	    System.out.println("FileBasedRepository.put(): failed");
	}
    }

    // ------------------------------------------------------------
    public OBJ_TYPE get(String key) {

	if (verbosity.isEnabled())
	    System.out.println("FileBasedRepository.get(" + key + ")");
	
	OBJ_TYPE obj = null;

	try {
	    obj = bufferizer_.fromFile(new File(path_, key));
	}
	catch (Exception e) {
	    System.out.println("FileBasedRepository.get("
			       + key + "): failed");
	}
	
	return obj;
    }
    
    // ------------------------------------------------------------
    public long getSize(String key) {

	if (verbosity.isEnabled())
	    System.out.println("FileBasedRepository.getSize(" + key + ")");
	
	long result = 0;

	File file = new File(path_, key);
	if (file.exists())
	    result = file.length();
	return result;
    }
    
    // ------------------------------------------------------------
    public Collection<OBJ_TYPE> getElements() {

	if (verbosity.isEnabled())
	    System.out.println("FileBasedRepository.getElements()");
	
	File[] files = path_.listFiles();

	Collection<OBJ_TYPE> result = new HashSet<OBJ_TYPE>();
	for (File file: files) {
	    try {
		OBJ_TYPE obj = bufferizer_.fromFile(file);
		result.add(obj);
	    }
	    catch (Exception e) {
		System.out.println("FileBasedRepository.getElements(): failed");
	    }
	}

	return result;
    }
    
    // ------------------------------------------------------------
    public Collection<String> getKeys() {

	if (verbosity.isEnabled())
	    System.out.println("FileBasedRepository.getKeys()");

	File[] files = path_.listFiles();

	Collection<String> result = new HashSet<String>();
	for (File file: files) {
	    result.add(file.getName());
	}
	
	return result;
    }
    
    // ------------------------------------------------------------
    public int size() {

	if (verbosity.isEnabled())
	    System.out.println("FileBasedRepository.size()");

	File[] files = path_.listFiles();

	return files.length;
    }
    
    
}
