/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.util;

import java.io.BufferedReader;
import java.io.InputStreamReader;

public class History<TYPE> {

	// ------------------------------------------------------------
	public static void main(String args[]) throws Exception {

		History<String> history = new History<String>(20);
		// history.verbosity.enable();

		BufferedReader reader = new BufferedReader(new InputStreamReader(
				System.in));

		String str = "help";
		while (!str.equals("quit")) {
			System.out.print("\n% ");
			str = reader.readLine();
			history.put(str);
			System.out.print(history);
		}
		str = "";
		while (!str.equals("quit")) {
			System.out.print("\n? ");
			str = reader.readLine();
			long now = System.currentTimeMillis();
			long first = history.lookupFirst(str);
			long last = history.lookupLast(str);
			System.out.print("String '" + str + "' ");
			if (first == -1)
				System.out.println("is not in history");
			else {
				if (first == last)
					System.out.println("was seen once "
							+ ((now - first) / 1000) + " seconds ago");
				else
					System.out.println("was first seen "
							+ ((now - first) / 1000)
							+ " seconds ago, and last seen "
							+ ((now - last) / 1000) + " seconds ago");
			}
		}
	}

	private int head_, tail_;

	private long[] recordingTimes_;

	private int size_;

	private Object[] values_;

	public Verbosity verbosity = new Verbosity();

	// ------------------------------------------------------------
	public History(int size) {
		size_ = size;
		values_ = new Object[size_];
		recordingTimes_ = new long[size_];
		head_ = 0;
		tail_ = 0;
	}

	// ------------------------------------------------------------
	public synchronized long lookupFirst(TYPE value) {

		long recorded = -1;

		int cursor = tail_;
		while ((cursor != head_) && (recorded == -1)) {
			if (value.equals(values_[cursor]))
				recorded = recordingTimes_[cursor];
			cursor = (cursor + 1) % size_;
		}

		if (verbosity.isEnabled()) {
			System.out.println("History.lookupFirst('" + value + "'): "
					+ recorded);
		}

		return recorded;
	}

	// ------------------------------------------------------------
	public synchronized long lookupLast(TYPE value) {

		if (verbosity.isEnabled()) {
			System.out.println("History.lookupLast('" + value + "')");
		}

		long recorded = -1;

		if (head_ != tail_) {
			int cursor = (head_ + size_ - 1) % size_;
			while ((cursor != (tail_ + size_ - 1) % size_) && (recorded == -1)) {
				if (value.equals(values_[cursor]))
					recorded = recordingTimes_[cursor];
				cursor = (cursor + size_ - 1) % size_;
			}
		}

		if (verbosity.isEnabled()) {
			System.out.println("History.lookupLast('" + value + "'): "
					+ recorded);
		}

		return recorded;
	}

	// ------------------------------------------------------------
	public synchronized void purgeAfter(long date) {

		if (verbosity.isEnabled()) {
			System.out.println("History.purgeAfter('" + date + "')");
		}

		while ((tail_ != head_) && (recordingTimes_[head_] > date))
			head_ = (head_ > 0 ? head_ - 1 : size_ - 1);
	}

	// ------------------------------------------------------------
	public synchronized void purgeBefore(long date) {

		if (verbosity.isEnabled()) {
			System.out.println("History.purgeBefore('" + date + "')");
		}

		while ((tail_ != head_) && (recordingTimes_[tail_] < date))
			tail_ = (tail_ + 1) % size_;
	}

	// ------------------------------------------------------------
	public synchronized void put(TYPE value) {

		if (verbosity.isEnabled())
			System.out.println("History.put('" + value + "')");

		long now = System.currentTimeMillis();

		// long recorded = lookup(value);

		// if (recorded != -1) {
		// if (verbosity.isEnabled())
		// System.out.println("Id '"
		// + value + "' was recorded "
		// + (now - recorded)/1000 + " seconds ago");
		// return;
		// }

		recordingTimes_[head_] = now;
		values_[head_] = value;
		head_ = (head_ + 1) % size_;
		if (head_ == tail_) {
			tail_ = (tail_ + 1) % size_;
			if (verbosity.isEnabled()) {
				System.out
						.println("History.put('"
								+ value
								+ "'): WARNING: a non-obsolete entry has been overwritten");
			}
		}

		if (verbosity.isEnabled()) {
			System.out.println("History.put('" + value + "') done");
			System.out.println(this);
		}
	}

	// ------------------------------------------------------------
	private synchronized String queueToString() {
		long now = System.currentTimeMillis();
		String result = "";
		int cursor = tail_;
		while (cursor != head_) {
			result += "['" + values_[cursor] + "', "
					+ ((now - recordingTimes_[cursor]) / 1000) + "] ";
			cursor = (cursor + 1) % size_;
		}
		return result;
	}

	// ------------------------------------------------------------
	@Override
	public String toString() {
		return "History(size= " + size_ + ", head= " + head_ + ", tail= "
				+ tail_ + ", queue= " + queueToString();
	}
}
