/* *****************************************************************************
 * Copyright (c) 2005-2010 VALORIA Laboratory, 
 * Universite Europeenne de Bretagne, Universite de Bretagne-Sud, France
 * <http://www-valoria.univ-ubs/CASA/DoDWAN>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************
 * $Id: Log2DGS.java $
 * ****************************************************************************/

package casa.dodwan.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.OutputStream;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;

import java.util.Arrays;
import java.util.Comparator;

/**
 * Parses a log file and detects all events pertaining to node
 * activity (i.e. online/offline events) as well as neighbor discovery
 * (i.e. add/remove neighbor). Produces a DGS file based on these
 * events.
 */
public class Log2DGSbis extends LogParser {

    private long lastEventTime_ = -1;
    private long lastTimeDisplayed_ = -1;
    
    private Map<String,Set<String>> nodes_ =
	new HashMap<String,Set<String>>();

    private Map<String,Long> lastSeen_ =
	new HashMap<String,Long>();

    private Set<Action> actions_ = new HashSet<Action>();

    // ==================================================
    public class Action {

	public long time;
	public String action;

	public Action(long time, String action) {
	    this.time = time;
	    this.action = action;
	}
    }
    // ==================================================
    private class ActionComparator
        implements Comparator<Action> {

        public int compare(Action a1, Action a2) {
            return (int)(a1.time - a2.time);
        }
    }
     // ==================================================
   
    // ------------------------------------------------------------
    public Log2DGSbis(InputStream in, OutputStream out) {
	super(in, out);
    }

    // ------------------------------------------------------------
    public void processEvent(Event event) {

	if (event.tag.equals("dodwan")
	    && event.message.startsWith("online")) {
	    processOnline(event);
	    return;
	}

	if (event.tag.equals("dodwan")
	    && event.message.startsWith("offline")) {
	    processOffline(event);
	    return;
	}

	if (event.tag.equals("tcp_chan")) {
		if (event.message.startsWith("handshake")) {
		    processHandshake(event);
		    return;
		}
		if (event.message.startsWith("< ")) {
		    processReceive(event);
		    return;
		}
		if (event.message.startsWith("open TCP_Channel")
		    || event.message.startsWith("accept TCP_Channel")) {
		    processInit(event);
		    return;
		}
	    }
    }

    // ------------------------------------------------------------
    private void displayTime(long now) {

	if (now != lastTimeDisplayed_) {
	    lastTimeDisplayed_ = now;
	    System.out.println("st " + now);
	}
    }
    
    // // ------------------------------------------------------------
    private boolean isOnline(String id) {

	Set<String> neigh = nodes_.get(id);
	return (neigh != null);
    }

    // ------------------------------------------------------------
    private void processOnline(Event event) {

	if (isOnline(event.name)) {
	    System.err.println("Error: node " + event.name + " is already online");
	    System.exit(1);
	}
    
	addNode(event.name);
	actions_.add(new Action(event.time, "an " + event.name));
	lastEventTime_ = event.time;
    }
    

    // ------------------------------------------------------------
    private void addNode(String id) {

	nodes_.put(id, new HashSet<String>());
    }

    // ------------------------------------------------------------
    private void processOffline(Event event) {

	if (! isOnline(event.name)) {
	    System.err.println("Error: node " + event.name + " is not online");
	    System.exit(1);
	}
	       
	deleteNode(event.time, event.name);
	actions_.add(new Action(event.time, "dn " + event.name));
	lastEventTime_ = event.time;
    }
    

    // ------------------------------------------------------------
    private void deleteNode(long time, String id) {

	Set<String> neigh = nodes_.get(id);
	if (! neigh.isEmpty()) {
	    // System.out.println("# Removing edges before node " + id + " is removed");
	    Set<String> tmp = new HashSet<String>();
	    tmp.addAll(neigh);
	    for (String id2: tmp) {
		deleteEdge(time, id, id2);
	    }
	}

	nodes_.remove(id);
    }

    // ------------------------------------------------------------
    private boolean areNeighbors(String node1, String node2) {

	Set<String> neigh1 = nodes_.get(node1);
	Set<String> neigh2 = nodes_.get(node2);

	if ((neigh1 == null) || (neigh2 == null))
	    return false;
	
	return (neigh1.contains(node2));
    }

    // ------------------------------------------------------------
    private void processHandshake(Event event) {

	int idx1 = event.message.indexOf("=");
	int idx2 = event.message.indexOf(")");
	String node1 = event.name;
	String node2 = event.message.substring(idx1+1, idx2);

	if (! (isOnline(node1) && isOnline(node2))) {
	    System.err.println("Error: handshake observed between nodes that are not online ("
			       + node1 + " and " + node2 + ")");
	    System.exit(1);
	}

	if (node1.compareTo(node2) >= 0)
	    return;
	// if (areNeighbors(node1, node2))
	//     return;

	addEdge(event.time, node1, node2);
	lastEventTime_ = event.time;
    }

    // ------------------------------------------------------------
    private String getEdgeId(String node1, String node2) {

	if (node1.compareTo(node2) < 0)
	    return node1 + "-" + node2;
	else
	    return node2 + "-" + node1;
    }

    // ------------------------------------------------------------
    private void addEdge(long time, String node1, String node2) {

	// if (areNeighbors(node1, node2))
	//     return;

	String edgeId = getEdgeId(node1, node2);
	if (lastSeen_.get(edgeId) != null)
	    return;
	
	actions_.add(new Action(time, "ae " + edgeId
				+ " " + node1 + " " + node2));
	lastSeen_.put(edgeId, time);
	
	Set<String> neigh = nodes_.get(node1);
	neigh.add(node2);
	neigh = nodes_.get(node2);
	neigh.add(node1);
    }

    // ------------------------------------------------------------
    private void processInit(Event event) {

	int idx1 = event.message.indexOf("=");
	idx1 = event.message.indexOf("=", idx1+1);
	int idx2 = event.message.indexOf(":");
	idx2 = event.message.indexOf(":", idx2+1);
	String node1 = event.name;
	String node2 = event.message.substring(idx1+1, idx2);

	String edgeId = getEdgeId(node1, node2);
	Long lastSeen = lastSeen_.get(edgeId);
	if (lastSeen != null) {
	    deleteEdge(lastSeen, node1, node2);
	    // actions_.add(new Action(lastSeen, "de " + edgeId));
	    // lastSeen_.remove(edgeId);
	}
    }
    
    // ------------------------------------------------------------
    private void processReceive(Event event) {

	int idx1 = event.message.indexOf("=");
	idx1 = event.message.indexOf("=", idx1+1);
	int idx2 = event.message.indexOf(":");
	idx2 = event.message.indexOf(":", idx2+1);
	String node1 = event.name;
	String node2 = event.message.substring(idx1+1, idx2);

	String edgeId = getEdgeId(node1, node2);
	if (lastSeen_.get(edgeId) == null)
	    return;
	
	lastSeen_.put(edgeId, event.time);
	
	lastEventTime_ = event.time;

    }

    // ------------------------------------------------------------
    // private void processRemoveNeighbor(Event event) {

    // 	String[] fields = event.message.split(" ");
    // 	String node1 = event.name;
    // 	String node2 = fields[1];

    // 	if (! areNeighbors(node1, node2))
    // 	    return;
	
    // 	deleteEdge(event.time, node1, node2);
    // }

    // ------------------------------------------------------------
    private void deleteEdge(long time, String node1, String node2) {

	String edgeId = getEdgeId(node1, node2);

	actions_.add(new Action(time, "de " + edgeId));

	lastSeen_.remove(edgeId);
	
	Set<String> neigh = nodes_.get(node1);
	neigh.remove(node2);
	neigh = nodes_.get(node2);
	neigh.remove(node1);
    }
    
    // ------------------------------------------------------------
    public void writeOutput() {

	Action[] array = new Action[actions_.size()];
        actions_.toArray(array);
        Arrays.sort(array, new ActionComparator());

	
	System.out.println("DGS004");
	System.out.println("null 0 0");
	System.out.println("cg ref_time=\"" + ref_time + "\"");
	long  st = -1;
	for (int i=0; i<array.length; i++) {
	    Action action = array[i];
	    if (action.time != st) {
		st = action.time;
		System.out.println("st " + st);
	    }
            System.out.println(action.action);
	}
    }

    // ------------------------------------------------------------
    private void close() {

	Set<String> tmp = new HashSet<String>();

	if (tmp.isEmpty())
	    return;
	
	System.out.println("# Removing all remaining nodes");
	tmp.addAll(nodes_.keySet());
	for (String id: tmp) {
	    deleteNode(lastEventTime_, id);
	}
    }

    // ------------------------------------------------------------
    public static void usage() {
        System.out.println("java Log2DGS [ref_time] [-|logfile]");
        System.out.println("Reads a log file on stdin and produces a DGS file based on log events");
	System.out.println("pertaining to node activity (i.e. online|offline) and neighbor");
	System.out.println("discovery (i.e. add/remove).");
        System.exit(0);
    }

   // ------------------------------------------------------------
    public static void main(String[] args) throws IOException {
        if ((args.length > 0 && args[0].equals("-h"))
	    || (args.length != 2)) {
            usage();
        }

	long ref_time = Long.parseLong(args[0]);
	
	InputStream is;
	if (args[1].equals("-"))
	    is = System.in;
	else
	    is = new FileInputStream(args[1]);
	
        Log2DGSbis parser =
	    new Log2DGSbis(is, System.out);
	parser.setRefTime(ref_time);
	
	parser.parse();
	parser.close();

	parser.writeOutput();
    }
}
